"""Context and settings for :mod:`message_ix_models` code."""
import logging
from copy import deepcopy
from dataclasses import fields
from pathlib import Path
from typing import List

import ixmp
import message_ix
from click import BadOptionUsage

from .config import Config

log = logging.getLogger(__name__)

#: List of Context instances, from first created to last.
_CONTEXTS: List["Context"] = []


# Configuration keys which can be accessed directly on Context.
_ALIAS = dict()
_ALIAS.update({f.name: "core" for f in fields(Config)})


def _dealiased(base: str, data: dict) -> dict:
    """Separate values from `data` which belong on `base` according to `_ALIAS`."""
    result = {}
    for name, path in filter(lambda ap: ap[1] == base, _ALIAS.items()):
        try:
            result[name] = data.pop(name)
        except KeyError:
            pass

    if len(result):
        log.warning(
            f"Create a Config instance instead of passing {list(result.keys())} to"
            " Context()"
        )

    return result


class Context(dict):
    """Context and settings for :mod:`message_ix_models` code."""

    # NB the docs contain a table of settings

    @classmethod
    def get_instance(cls, index=0) -> "Context":
        """Return a Context instance; by default, the first created.

        Parameters
        ----------
        index : int, *optional*
            Index of the Context instance to return, e.g. ``-1`` for the most recently
            created.
        """
        return _CONTEXTS[index]

    @classmethod
    def only(cls) -> "Context":
        """Return the only :class:`.Context` instance.

        Raises
        ------
        IndexError
            If there is more than one instance.
        """
        if len(_CONTEXTS) > 1:
            raise IndexError(f"ambiguous: {len(_CONTEXTS)} Context instances")
        return _CONTEXTS[0]

    def __init__(self, *args, **kwargs):
        from message_ix_models.model import Config as ModelConfig
        from message_ix_models.report import Config as ReportConfig

        if len(_CONTEXTS) == 0:
            log.info("Create root Context")

        # Handle keyword arguments going to known config dataclasses
        kwargs["core"] = Config(**_dealiased("core", kwargs))
        kwargs["model"] = ModelConfig(**_dealiased("model", kwargs))
        kwargs["report"] = ReportConfig()

        # Store any keyword arguments
        super().__init__(*args, **kwargs)

        # Store a reference for get_instance()
        _CONTEXTS.append(self)

    def _dealias(self, name):
        base = _ALIAS[name]

        # Warn about direct reference to aliased attributes
        if base not in {"core", "model"}:  # pragma: no cover
            log.warnings(f"Use Context.{base}.{name} instead of Context.{name}")

        return getattr(self, base), name

    # Item access
    def __getitem__(self, name):
        try:
            return getattr(*self._dealias(name))
        except KeyError:
            return super().__getitem__(name)

    def __setitem__(self, name, value):
        try:
            return setattr(*self._dealias(name), value)
        except KeyError:
            super().__setitem__(name, value)

    def update(self, arg=None, **kwargs):
        # Force update() to use the __setitem__ above
        for k, v in dict(*filter(None, [arg]), **kwargs).items():
            self.__setitem__(k, v)

    # Attribute access
    def __setattr__(self, name, value):
        self[name] = value

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name) from None

    def __deepcopy__(self, memo):
        mp = self.pop("_mp", None)

        result = deepcopy(super(), memo)

        if mp is not None:
            self._mp = mp

        _CONTEXTS.append(result)

        return result

    def __repr__(self):
        return f"<{self.__class__.__name__} object at {id(self)} with {len(self)} keys>"

    def delete(self):
        """Hide the current Context from future :meth:`.get_instance` calls."""
        # Index of the *last* matching instance
        index = len(_CONTEXTS) - 1 - list(reversed(_CONTEXTS)).index(self)

        if index > 0:
            _CONTEXTS.pop(index)
        else:  # pragma: no cover
            # The `session_context` fixture means this won't occur during tests
            log.warning("Won't delete the only Context instance")

        self.close_db()

    def write_debug_archive(self) -> None:
        """Write an archive containing the files listed in :attr:`.debug_paths`.

        The archive file name is constructed using :func:`.unique_id` and appears in a
        :file:`debug` subdirectory under the :ref:`local data path <local-data>`.

        The archive also contains a file :file:`command.txt` that gives the full
        command-line used to invoke :program:`mix-models`.
        """
        from zipfile import ZIP_DEFLATED, ZipFile

        from .click import format_sys_argv, unique_id

        # Output file
        target = self.core.local_data.joinpath("debug", f"{unique_id()}.zip")
        log.info(f"Write to: {target}")

        target.parent.mkdir(parents=True, exist_ok=True)

        with ZipFile(target, mode="w", compression=ZIP_DEFLATED) as zf:
            # Write a file that contains the CLI invocation
            zf.writestr("command.txt", format_sys_argv())

            # Write the identified files
            for dp in self.core.debug_paths:
                if not dp.exists():
                    log.info(f"Not found: {dp}")
                    continue

                zf.write(dp, arcname=dp.relative_to(self.core.local_data))
                # log.info(debug_path)

    def clone_to_dest(self, create=True) -> message_ix.Scenario:
        """Return a scenario based on the ``--dest`` command-line option.

        Parameters
        ----------
        create : bool, *optional*
            If :obj:`True` (the default) and the base scenario does not exist, a bare
            RES scenario is created. Otherwise, an exception is raised.

        Returns
        -------
        Scenario
            To prevent the scenario from being garbage collected, keep a reference to
            its Platform:

            .. code-block: python

               s = context.clone_to_dest()
               mp = s.platform

        See also
        --------
        create_res
        """
        cfg = self.core
        if not cfg.dest_scenario:
            # No information on the destination; try to parse a URL, storing the keys
            # dest_platform and dest_scenario.
            self.handle_cli_args(
                url=cfg.dest, _store_as=("dest_platform", "dest_scenario")
            )

        try:
            # Get the base scenario, e.g. from the --url CLI argument
            scenario_base = self.get_scenario()

            # By default, clone to the same platform
            mp_dest = scenario_base.platform

            try:
                if cfg.dest_platform["name"] != mp_dest.name:
                    # Different platform
                    # Not tested; current test fixtures make it difficult to create
                    # *two* temporary platforms simultaneously
                    mp_dest = ixmp.Platform(**cfg.dest_platform)  # pragma: no cover
            except KeyError:
                pass
        except Exception as e:
            log.info("Base scenario not given or found")
            log.debug(f"{type(e).__name__}: {e}")

            if not create:
                log.error("and create=False")
                raise

            # Create a bare RES to be the base scenario

            from message_ix_models.model.bare import create_res

            # Create on the destination platform
            ctx = deepcopy(self)
            ctx.core.platform_info.update(cfg.dest_platform)
            scenario_base = create_res(ctx)

            # Clone to the same platform
            mp_dest = scenario_base.platform

        # Clone
        log.info(f"Clone to {repr(cfg.dest_scenario)}")
        return scenario_base.clone(platform=mp_dest, **cfg.dest_scenario)

    def close_db(self):
        try:
            mp = self.pop("_mp")
            mp.close_db()
        except KeyError:
            pass

    def get_cache_path(self, *parts) -> Path:
        """Return a path to a local cache file, i.e. within :attr:`.Config.cache_path`.

        The directory containing the resulting path is created if it does not already
        exist.
        """
        # Construct relative to local_data if cache_path is not defined
        base = self.core.cache_path or self.core.local_data.joinpath("cache")

        result = base.joinpath(*parts)

        # Ensure the directory exists
        result.parent.mkdir(parents=True, exist_ok=True)

        return result

    def get_local_path(self, *parts: str, suffix=None) -> Path:
        """Return a path under :attr:`.Config.local_data`.

        Parameters
        ==========
        parts :
            Path fragments, for instance directories, passed to
            :meth:`~.pathlib.PurePath.joinpath`.
        suffix :
            File name suffix including a "."—for instance, ".csv"—passed to
            :meth:`~.pathlib.PurePath.with_suffix`.
        """
        result = self.core.local_data.joinpath(*parts)
        return result.with_suffix(suffix) if suffix else result

    def get_platform(self, reload=False) -> ixmp.Platform:
        """Return a |Platform| from :attr:`.Config.platform_info`.

        When used through the CLI, :attr:`.Config.platform_info` is a 'base' platform as
        indicated by the --url or --platform  options.

        If a Platform has previously been instantiated with :meth:`get_platform`, the
        same object is returned unless `reload=True`.
        """
        if not reload:
            # Return an existing Platform, if any
            try:
                return self["_mp"]
            except KeyError:
                pass

        # Close any existing Platform, e.g. to reload it
        try:
            self["_mp"].close_db()
            del self["_mp"]
        except KeyError:
            pass

        # Create a Platform
        self["_mp"] = ixmp.Platform(**self.core.platform_info)
        return self["_mp"]

    def get_scenario(self) -> message_ix.Scenario:
        """Return a |Scenario| from :attr:`~.Config.scenario_info`.

        When used through the CLI, :attr:`~.Config.scenario_info` is a ‘base’ scenario
        for an operation, indicated by the ``--url`` or
        ``--platform/--model/--scenario`` options.
        """
        return message_ix.Scenario(self.get_platform(), **self.core.scenario_info)

    def set_scenario(self, scenario: message_ix.Scenario) -> None:
        """Update :attr:`.Config.scenario_info` to match an existing `scenario`."""
        self.core.scenario_info.update(
            model=scenario.model, scenario=scenario.scenario, version=scenario.version
        )

    def handle_cli_args(
        self,
        url=None,
        platform=None,
        model_name=None,
        scenario_name=None,
        version=None,
        local_data=None,
        verbose=False,
        _store_as=("platform_info", "scenario_info"),
    ):
        """Handle command-line arguments.

        May update the :attr:`.Config.local_data`, :attr:`~.Config.platform_info`,
        :attr:`~.Config.scenario_info`, and/or :attr:`~.Config.url` settings.
        """
        self.core.verbose = verbose

        # Store the path to command-specific data and metadata
        if local_data:
            self.core.local_data = local_data

        # References to the Context settings to be updated
        platform_info = getattr(self.core, _store_as[0])
        scenario_info = getattr(self.core, _store_as[1])

        # Store information for the target Platform
        if url:
            if platform or model_name or scenario_name or version:
                raise BadOptionUsage(
                    "--platform --model --scenario and/or --version",
                    " redundant with --url",
                )

            self.core.url = url
            urlinfo = ixmp.utils.parse_url(url)
            platform_info.update(urlinfo[0])
            scenario_info.update(urlinfo[1])
        elif platform:
            platform_info["name"] = platform

        # Store information about the target Scenario
        if model_name:
            scenario_info["model"] = model_name
        if scenario_name:
            scenario_info["scenario"] = scenario_name
        if version:
            scenario_info["version"] = version

    def use_defaults(self, settings):
        """Update from `settings`."""
        for setting, info in settings.items():
            if setting not in self:
                log.info(f"Use default {setting}={info[0]}")

            value = self.setdefault(setting, info[0])

            if value not in info:
                raise ValueError(f"{setting} must be in {info}; got {value}")



import message_ix
import pytest

from message_ix_models import testing
from message_ix_models.model.bare import Config

#: Number of items in the respective YAML files.
SET_SIZE = dict(
    commodity=18,
    level=6,
    node=14 + 1,  # R14 is default, and 'World' exists automatically
    relation=20,
    technology=377,
    year=28,  # YB is default: 1950, ..., 2020, 2025, ..., 2055, 2060, ..., 2110
)


@pytest.mark.parametrize(
    "settings, expected",
    [
        # Defaults per bare.SETTINGS
        (dict(), dict()),
        #
        # Different regional aggregations
        (dict(regions="R11"), dict(node=11 + 1)),
        (dict(regions="RCP"), dict(node=5 + 1)),
        # MESSAGE-IL
        (dict(regions="ISR"), dict(node=1 + 1)),
        #
        # Different time periods
        (dict(years="A"), dict(year=16)),  # ..., 2010, 2020, ..., 2110
        #
        # Option to add a dummy technology/commodity so the model solves
        (
            dict(res_with_dummies=True),
            dict(
                commodity=SET_SIZE["commodity"] + 1,
                technology=SET_SIZE["technology"] + 2,
            ),
        ),
    ],
)
def test_create_res(request, test_context, settings, expected):
    # Apply settings to the temporary context
    test_context.model = Config(**settings)

    # Call bare.create_res() via testing.bare_res(). This ensures the slow step of
    # creating the scenario occurs only once per test session. If it fails, it will
    # either fail within this test, or in some other test function that calls
    # testing.bare_res() with the same arguments.
    scenario = testing.bare_res(request, test_context, solved=False)

    # Returns a Scenario object
    assert isinstance(scenario, message_ix.Scenario)

    # Sets contain the expected number of elements
    sets = SET_SIZE.copy()
    sets.update(expected)
    for name, size in sets.items():
        values = scenario.set(name)
        assert size == len(values), (name, values)


class TestConfig:
    def test_init(self):
        c = Config()
        c.regions = "R999"

        # TODO expand



"""Tests of :mod:`message_ix_models.util.node`."""
import re

import pandas as pd
import pytest
from genno import Quantity
from message_ix import Scenario, make_df

from message_ix_models.model.bare import create_res
from message_ix_models.model.structure import get_codes
from message_ix_models.util import (
    adapt_R11_R12,
    adapt_R11_R14,
    broadcast,
    identify_nodes,
)
from message_ix_models.util.node import MappingAdapter


def test_mapping_adapter():
    """Generic test of MappingAdapter."""
    a = MappingAdapter({"foo": [("a", "x"), ("a", "y"), ("b", "z")]})

    columns = ["foo", "bar", "value"]

    df = pd.DataFrame([["a", "m", 1], ["b", "n", 2]], columns=columns)

    result = a(df)

    assert all(columns + ["unit"] == result.columns)

    with pytest.raises(TypeError):
        a(1.2)


PAR = "technical_lifetime"
VALUE = [0.1, 0.2]


@pytest.fixture(scope="function")
def input():
    """Fixture: test data for :obj:`.adapt_R11_R14`."""
    R11_all = get_codes("node/R11")
    R11_reg = R11_all[R11_all.index("World")].child
    df = make_df(
        PAR, technology="coal_ppl", year_vtg=[2021, 2022], value=[1.2, 3.4], unit="year"
    ).pipe(broadcast, node_loc=R11_reg)

    # Set a specific value for the regions to be broadcast
    df["value"] = df["value"].where(df["node_loc"] != "R11_CPA", VALUE[0])
    df["value"] = df["value"].where(df["node_loc"] != "R11_FSU", VALUE[1])

    return {PAR: df}


@pytest.mark.parametrize(
    "func,N,expected,target_nodes",
    [
        (adapt_R11_R12, 12, VALUE[0], ("R12_CHN", "R12_RCPA")),
        (adapt_R11_R14, 14, VALUE[1], ("R14_CAS", "R14_RUS", "R14_SCS", "R14_UBM")),
    ],
)
def test_adapt_df(input, func, N, expected, target_nodes):
    """:obj:`.adapt_R11_R14` handles :class:`pandas.DataFrame`."""
    # Function runs
    output = func(input)

    # Output is a dict containing 1 entry
    df_out = output.pop(PAR)
    assert 0 == len(output)

    # Output covers all R14 regions
    all_nodes = get_codes(f"node/R{N}")
    regions = all_nodes[all_nodes.index("World")].child
    assert set(regions) == set(df_out["node_loc"])

    # Output has expected length
    assert N * 2 == len(df_out)

    # Output values for new regions match input value for the base region
    assert (expected == df_out[df_out["node_loc"].isin(target_nodes)]["value"]).all()


@pytest.mark.parametrize(
    "func,expected,node_loc",
    [(adapt_R11_R12, VALUE[0], "R12_CHN"), (adapt_R11_R14, VALUE[1], "R14_CAS")],
)
def test_adapt_qty(input, func, expected, node_loc):
    """:obj:`.adapt_R11_R14` handles :class:`genno.Quantity`."""
    # Convert to genno.Quantity
    df = input[PAR]
    input[PAR] = Quantity.from_series(df.set_index(df.columns[:-2].tolist())["value"])
    input[PAR].attrs["_unit"] = df["unit"].unique()[0]

    # Function runs
    output = func(input)

    # Output is a dict containing 1 entry
    qty_out = output.pop(PAR)
    assert 0 == len(output)

    assert isinstance(qty_out, Quantity)
    assert "year" == qty_out.attrs["_unit"]

    # Output values for new regions match input value for the base region
    assert (
        expected == qty_out.sel(node_loc=node_loc, technology="coal_ppl", year=2022)
    ).all()


@pytest.mark.parametrize("regions", ["R11", "R12", "R14"])
def test_identify_nodes(caplog, test_context, regions):
    ctx = test_context
    ctx.regions = regions
    scenario = create_res(ctx)

    # The ID of the node codelist can be recovered from scenario contents
    assert regions == identify_nodes(scenario)

    # A node like "R11_GLB" is ignored
    with scenario.transact():
        scenario.add_set("node", f"{regions}_GLB")
    assert regions == identify_nodes(scenario)

    # Remove one element from the node set
    with scenario.transact():
        scenario.remove_set("node", scenario.set("node").tolist()[0])

    # No longer any match
    with pytest.raises(
        ValueError, match=f"IDs suggest codelist {repr(regions)}, values do not match"
    ):
        identify_nodes(scenario)


def test_identify_nodes1(test_context):
    mp = test_context.get_platform()
    scenario = Scenario(
        mp, model="identify_nodes", scenario="identify_nodes", version="new"
    )
    scenario.add_set("technology", "t")
    scenario.add_set("year", 0)
    scenario.commit("")

    with scenario.transact():
        scenario.add_set("node", "R99_ZZZ")

    with pytest.raises(
        ValueError,
        match=re.escape("Couldn't identify node codelist from ['R99_ZZZ', 'World']"),
    ):
        identify_nodes(scenario)



"""Utilities for nodes."""
import logging
from typing import List, Sequence, Union

from message_ix import Scenario
from sdmx.model.v21 import Code

from .common import Adapter, MappingAdapter  # noqa: F401

log = logging.getLogger(__name__)

#: Names of dimensions indexed by 'node'.
#:
#: .. todo:: to be robust to changes in :mod:`message_ix`, read these names from that
#:    package.
NODE_DIMS = [
    "n",
    "node",
    "node_loc",
    "node_origin",
    "node_dest",
    "node_rel",
    "node_share",
]

#: Mapping from R11 to R12 node IDs.
R11_R12 = (
    ("R11_AFR", "R12_AFR"),
    ("R11_CPA", "R12_CHN"),
    ("R11_EEU", "R12_EEU"),
    ("R11_FSU", "R12_FSU"),
    ("R11_LAM", "R12_LAM"),
    ("R11_MEA", "R12_MEA"),
    ("R11_NAM", "R12_NAM"),
    ("R11_PAO", "R12_PAO"),
    ("R11_PAS", "R12_PAS"),
    ("R11_CPA", "R12_RCPA"),
    ("R11_SAS", "R12_SAS"),
    ("R11_WEU", "R12_WEU"),
)

#: Mapping from R11 to R14 node IDs.
R11_R14 = (
    ("R11_AFR", "R14_AFR"),
    ("R11_FSU", "R14_CAS"),
    ("R11_CPA", "R14_CPA"),
    ("R11_EEU", "R14_EEU"),
    ("R11_LAM", "R14_LAM"),
    ("R11_MEA", "R14_MEA"),
    ("R11_NAM", "R14_NAM"),
    ("R11_PAO", "R14_PAO"),
    ("R11_PAS", "R14_PAS"),
    ("R11_FSU", "R14_RUS"),
    ("R11_SAS", "R14_SAS"),
    ("R11_FSU", "R14_SCS"),
    ("R11_FSU", "R14_UBM"),
    ("R11_WEU", "R14_WEU"),
)

#: Adapt data from the R11 to the R14 node list.
#:
#: The data is adapted using the mappings in :data:`R11_R12` for each of the dimensions
#: in :data:`NODE_DIMS`.
adapt_R11_R12 = MappingAdapter({d: R11_R12 for d in NODE_DIMS})

#: Adapt data from the R11 to the R14 node list.
#:
#: The data is adapted using the mappings in :data:`R11_R14` for each of the dimensions
#: in :data:`NODE_DIMS`.
adapt_R11_R14 = MappingAdapter({d: R11_R14 for d in NODE_DIMS})


def identify_nodes(scenario: Scenario) -> str:
    """Return the ID of a node codelist given the contents of `scenario`.

    Returns
    -------
    str
        The ID of the :doc:`/pkg-data/node` containing the regions of `scenario`.

    Raises
    ------
    ValueError
        if no codelist can be identified, or the nodes in the scenario do not match the
        children of the “World” node in the codelist.
    """
    from message_ix_models.model.structure import get_codes

    nodes = sorted(scenario.set("node"))

    # Candidate ID: split e.g. "R14_AFR" to "R14"
    id = nodes[0].split("_")[0]

    try:
        # Get the corresponding codelist
        codes = get_codes(f"node/{id}")
    except FileNotFoundError:
        raise ValueError(f"Couldn't identify node codelist from {repr(nodes)}")

    glb_node = [n.endswith("_GLB") for n in nodes]
    if any(glb_node):
        omit = nodes.pop(glb_node.index(True))
        log.info(f"Omit known, non-standard node '{omit}' from set to match")

    # Expected list of nodes
    world = codes[codes.index("World")]  # type: ignore [arg-type]
    codes = [world] + world.child

    try:
        assert set(nodes) == set(map(str, codes))
    except AssertionError:
        raise ValueError(
            "\n".join(
                [
                    f"Node IDs suggest codelist {repr(id)}, values do not match:",
                    repr(nodes),
                    repr(codes),
                ]
            )
        )
    else:
        log.info(f"Identified node codelist {repr(id)}")
        return id


def nodes_ex_world(nodes: Sequence[Union[str, Code]]) -> List[Union[str, Code]]:
    """Exclude "World" and anything containing "GLB" from `nodes`.

    May also be used as a reporting computation.

    .. todo:: Make available from :mod:`message_ix_models.report`.
    """
    return list(filter(lambda n_: "GLB" not in n_ and n_ != "World", nodes))



"""Command-line interface for MESSAGEix-GLOBIOM model tools.

Every tool and script in this repository is accessible through this CLI. Scripts are
grouped into commands and sub-commands. For help on specific (sub)commands, use --help,
e.g.:

    \b
    mix-models cd-links --help
    mix-models cd-links run --help

The top-level options --platform, --model, and --scenario are used by commands that
access specific message_ix scenarios; these can also be specified with --url.

For more information, see https://docs.messageix.org/projects/models2/en/latest/cli.html
"""
import logging
import sys
from pathlib import Path

import click
from ixmp.cli import main as ixmp_cli

from message_ix_models.util._logging import mark_time
from message_ix_models.util._logging import setup as setup_logging
from message_ix_models.util.click import common_params
from message_ix_models.util.context import Context

log = logging.getLogger(__name__)


# Main command group. The code in this function is ALWAYS executed, so it should only
# include tasks that are common to all CLI commands.
@click.group(help=__doc__)
@click.option(
    "--url", metavar="ixmp://PLATFORM/MODEL/SCENARIO[#VERSION]", help="Scenario URL."
)
@click.option("--platform", metavar="PLATFORM", help="Configured platform name.")
@click.option(
    "--model", "model_name", metavar="MODEL", help="Model name for some commands."
)
@click.option(
    "--scenario",
    "scenario_name",
    metavar="SCENARIO",
    help="Scenario name for some commands.",
)
@click.option("--version", type=int, help="Scenario version.")
@click.option("--local-data", type=Path, help="Base path for local data.")
@common_params("verbose")
@click.pass_context
def main(click_ctx, **kwargs):
    # Start timer
    mark_time(quiet=True)

    # Log to console
    setup_logging(level="DEBUG" if kwargs.pop("verbose") else "INFO", console=True)

    # Store the most recently created instance of message_ix_models.Context. click
    # carries this object to any subcommand decorated with @click.pass_obj.
    # NB this can't be Context.only(). When click.testing.CliRunner is used, there may
    #    already be ≥2 Context instances created elsewhere in the test session before
    #    this function is called to run CLI commands within the test session.
    click_ctx.obj = Context.get_instance(-1)

    # Handle command-line parameters
    click_ctx.obj.handle_cli_args(**kwargs)

    # Close any database connections when the CLI exits
    click_ctx.call_on_close(click_ctx.obj.close_db)


@main.command("export-test-data")
@click.option("--exclude", default="", help="Sheets to exclude.")
@click.option("--nodes", default="R11_AFR,R11_CPA", help="Nodes to include.")
@click.option("--techs", default="coal_ppl", help="Technologies to include.")
@click.pass_obj
def export_test_data_cmd(ctx, exclude, nodes, techs):
    """Prepare data for testing.

    Option values for --exclude, --nodes, and --techs must be comma-separated lists.
    """
    from message_ix_models.testing import export_test_data

    # Store CLI options on the Context
    ctx.export_exclude = list(filter(None, exclude.split(",")))  # Exclude empty string
    ctx.export_nodes = nodes.split(",")
    ctx.export_techs = techs.split(",")

    mark_time()

    # Export
    export_test_data(ctx)

    mark_time()


@main.command(hidden=True)
@click.pass_obj
def debug(ctx):
    """Hidden command for debugging."""
    # Print the local data path
    log.debug(ctx.local_data)


# Attach the ixmp "config" CLI
main.add_command(ixmp_cli.commands["config"])


#: List of submodules providing CLI (sub)commands accessible through `mix-models`.
#: Each of these should contain a function named ``cli`` decorated with @click.command
#: or @click.group.
submodules = [
    "message_ix_models.model.cli",
    "message_ix_models.model.structure",
    "message_ix_models.model.water.cli",
    "message_ix_models.project.ssp",
    "message_ix_models.report.cli",
    "message_ix_models.testing.cli",
    "message_ix_models.util.pooch",
]

try:
    import message_data.cli
except ImportError:
    # message_data is not installed or contains some ImportError of its own
    from traceback import format_exception

    # Display information for debugging
    etype, value, tb = sys.exc_info()
    print("", *format_exception(etype, value, tb, limit=-1, chain=False)[1:], sep="\n")
else:  # pragma: no cover  (needs message_data)
    # Also add message_data submodules
    submodules.extend(
        f"message_data.{name}" for name in message_data.cli.MODULES_WITH_CLI
    )

for name in submodules:
    # Import the module and retrieve the click.Command object
    __import__(name)
    cmd = getattr(sys.modules[name], "cli")

    # Avoid replacing message-ix-models CLI with message_data CLI
    if cmd.name in main.commands:  # pragma: no cover  (needs message_data)
        log.warning(f"Skip {repr(cmd.name)} CLI from {repr(name)}; already defined")
        continue

    main.add_command(cmd)



from typing import IO, Union

import click


@click.group("testing")
def cli():
    """Manipulate test data."""
    pass


FILENAMES = [
    "advance/advance_compare_20171018-134445.csv.zip",
    "iea/372f7e29-en.zip",
    "iea/8624f431-en.zip",
    "iea/cac5fa90-en.zip",
    "ssp/SSP-Review-Phase-1.csv.gz",
    "ssp/SspDb_country_data_2013-06-12.csv.zip",
]


@cli.command("fuzz-private-data")
@click.option("--frac", type=float, default=1.0, help="Fraction of rows (default: all)")
@click.argument("filename", metavar="FILENAME", type=click.Choice(FILENAMES))
def fuzz_private_data(filename, frac: float):  # pragma: no cover
    """Create random data for testing.

    This command creates data files in message_ix_models/data/test/… based on
    corresponding private files in message_data/data/…. This supports testing of code in
    message_ix_models that handles these files.

    The files are identical in structure and layout, except the values are "fuzzed", or
    replaced with random values.

    To see valid FILENAMES, run the command with no arguments.
    """
    import zipfile
    from pathlib import Path
    from tempfile import TemporaryDirectory

    import dask.dataframe as dd
    import pandas as pd
    from numpy import char, random

    from message_ix_models.project.advance.data import NAME
    from message_ix_models.util import package_data_path, private_data_path

    # Paths
    p = Path(filename)
    path_in = private_data_path(p)
    path_out = package_data_path("test", p)

    # Read the data
    with TemporaryDirectory() as td:
        td_path = Path(td)
        if "advance" in filename:
            # Manually unpack one member of the multi-member archive `path_in`
            target: Union[IO, Path, str] = zipfile.ZipFile(path_in).extract(
                NAME, path=td_path
            )
        elif "iea" in filename:
            # Manually unpack so that dask.dataframe.read_csv() can be used
            from message_ix_models.tools.iea.web import unpack_zip

            target = unpack_zip(path_in)
        else:
            target = path_in

        # - Read the data using dask & pyarrow.
        # - Subset the data if `frac` < 1.0.
        # - Compute the resulting pandas.DataFrame
        df = dd.read_csv(target, engine="pyarrow").sample(frac=frac).compute()

    # Determine columns in which to replace numerical data
    if "iea" in filename:
        # Specific column
        cols = ["Value"]
    else:
        # All columns with numeric names, for instance 2000, 2001, etc.
        cols = list(filter(char.isnumeric, df.columns))

    # Shape of random data
    size = (df.shape[0], len(cols))
    # - Generate random data of this shape.
    # - Keep only the elements corresponding to non-NA elements of `df`.
    # - Update `df` with these values.*
    generator = random.default_rng()
    df.update(df.where(df.isna(), pd.DataFrame(generator.random(size), columns=cols)))

    # Write to file, keeping only a few decimal points
    path_out.parent.mkdir(parents=True, exist_ok=True)

    if "advance" in filename:
        zf = zipfile.ZipFile(path_out, "w", compression=zipfile.ZIP_BZIP2)
        target = zf.open(NAME)
    else:
        target = path_out

    df.to_csv(target, index=False, float_format="%.2f")



"""Prepare data for water use for cooling & energy technologies."""

import numpy as np
import pandas as pd
from message_ix import make_df

from message_ix_models.model.water.data.water_supply import map_basin_region_wat
from message_ix_models.util import (
    broadcast,
    make_matched_dfs,
    package_data_path,
    same_node,
)


# water & electricity for cooling technologies
def cool_tech(context):  # noqa: C901
    """Process cooling technology data for a scenario instance.
    The input values of parent technologies are read in from a scenario instance and
    then cooling fractions are calculated by using the data from
    ``tech_water_performance_ssp_msg.csv``.
    It adds cooling  technologies as addons to the parent technologies.The
    nomenclature for cooling technology is <parenttechnologyname>__<coolingtype>.
    E.g: `coal_ppl__ot_fresh`
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'.
        Values are data frames ready for :meth:`~.Scenario.add_par`.
        Years in the data include the model horizon indicated by
        ``context["water build info"]``, plus the additional year 2010.
    """
    # TODO reduce complexity of this function from 18 to 15 or less
    #: Name of the input file.
    # The input file mentions water withdrawals and emission heating fractions for
    # cooling technologies alongwith parent technologies:
    FILE = "tech_water_performance_ssp_msg.csv"
    # Investment costs & regional shares of hist. activities of cooling
    # technologies
    FILE1 = (
        "cooltech_cost_and_shares_"
        + (f"ssp_msg_{context.regions}" if context.type_reg == "global" else "country")
        + ".csv"
    )

    # define an empty dictionary
    results = {}

    # Reference to the water configuration
    info = context["water build info"]
    sub_time = context.time

    # reading basin_delineation
    FILE2 = f"basins_by_region_simpl_{context.regions}.csv"
    PATH = package_data_path("water", "delineation", FILE2)

    df_node = pd.read_csv(PATH)
    # Assigning proper nomenclature
    df_node["node"] = "B" + df_node["BCU_name"].astype(str)
    df_node["mode"] = "M" + df_node["BCU_name"].astype(str)
    if context.type_reg == "country":
        df_node["region"] = context.map_ISO_c[context.regions]
    else:
        df_node["region"] = f"{context.regions}_" + df_node["REGION"].astype(str)

    node_region = df_node["region"].unique()
    # reading ppl cooling tech dataframe
    path = package_data_path("water", "ppl_cooling_tech", FILE)
    df = pd.read_csv(path)
    cooling_df = df.loc[df["technology_group"] == "cooling"]
    # Separate a column for parent technologies of respective cooling
    # techs
    cooling_df["parent_tech"] = (
        cooling_df["technology_name"]
        .apply(lambda x: pd.Series(str(x).split("__")))
        .drop(columns=1)
    )

    scen = context.get_scenario()

    # Extracting input database from scenario for parent technologies
    # Extracting input values from scenario
    ref_input = scen.par("input", {"technology": cooling_df["parent_tech"]})
    # Extracting historical activity from scenario
    ref_hist_act = scen.par(
        "historical_activity", {"technology": cooling_df["parent_tech"]}
    )
    # Extracting historical capacity from scenario
    ref_hist_cap = scen.par(
        "historical_new_capacity", {"technology": cooling_df["parent_tech"]}
    )
    # cooling fraction = H_cool = Hi - 1 - Hi*(h_fg)
    # where h_fg (flue gasses losses) = 0.1
    ref_input["cooling_fraction"] = ref_input["value"] * 0.9 - 1

    def missing_tech(x):
        """Assign values to missing data.
        It goes through the input data frame and extract the technologies which
        don't have input values and then assign manual  values to those technologies
        along with assigning them an arbitrary level i.e dummy supply
        """
        data_dic = {
            "geo_hpl": 1 / 0.850,
            "geo_ppl": 1 / 0.385,
            "nuc_hc": 1 / 0.326,
            "nuc_lc": 1 / 0.326,
            "solar_th_ppl": 1 / 0.385,
        }

        if data_dic.get(x["technology"]):
            if x["level"] == "cooling":
                return pd.Series((data_dic.get(x["technology"]), "dummy_supply"))
            else:
                return pd.Series((data_dic.get(x["technology"]), x["level"]))
        else:
            return pd.Series((x["value"], x["level"]))

    ref_input[["value", "level"]] = ref_input[["technology", "value", "level"]].apply(
        missing_tech, axis=1
    )

    # Combines the input df of parent_tech with water withdrawal data
    input_cool = (
        cooling_df.set_index("parent_tech")
        .combine_first(ref_input.set_index("technology"))
        .reset_index()
    )

    # Drops NA values from the value column
    input_cool = input_cool.dropna(subset=["value"])

    # Convert year values into integers to be compatibel for model
    input_cool.year_vtg = input_cool.year_vtg.astype(int)
    input_cool.year_act = input_cool.year_act.astype(int)
    # Drops extra technologies from the data
    input_cool = input_cool[
        (input_cool["level"] != "water_supply") & (input_cool["level"] != "cooling")
    ]

    input_cool = input_cool[
        ~input_cool["technology_name"].str.contains("hpl", na=False)
    ]
    input_cool = input_cool[
        (input_cool["node_loc"] != f"{context.regions}_GLB")
        & (input_cool["node_origin"] != f"{context.regions}_GLB")
    ]

    def cooling_fr(x):
        """Calculate cooling fraction
        Returns the calculated cooling fraction after for two categories;
        1. Technologies that produce heat as an output
            cooling_fraction(h_cool) = input value(hi) - 1
        Simply subtract 1 from the heating value since the rest of the part is already
        accounted in the heating value
        2. Rest of technologies
            h_cool  =  hi -Hi* h_fg - 1,
            where:
                h_fg (flue gasses losses) = 0.1 (10% assumed losses)
        """
        if "hpl" in x["index"]:
            return x["value"] - 1
        else:
            return x["value"] - (x["value"] * 0.1) - 1

    input_cool["cooling_fraction"] = input_cool.apply(cooling_fr, axis=1)

    # Converting water withdrawal units to Km3/GWa
    # this refers to activity per cooling requirement (heat)
    input_cool["value_cool"] = (
        input_cool["water_withdrawal_mid_m3_per_output"]
        * 60
        * 60
        * 24
        * 365
        * 1e-9
        / input_cool["cooling_fraction"]
    )

    input_cool["return_rate"] = 1 - (
        input_cool["water_consumption_mid_m3_per_output"]
        / input_cool["water_withdrawal_mid_m3_per_output"]
    )
    # consumption to be saved in emissions rates for reporting purposes
    input_cool["consumption_rate"] = (
        input_cool["water_consumption_mid_m3_per_output"]
        / input_cool["water_withdrawal_mid_m3_per_output"]
    )

    input_cool["value_return"] = input_cool["return_rate"] * input_cool["value_cool"]

    # only for reporting purposes
    input_cool["value_consumption"] = (
        input_cool["consumption_rate"] * input_cool["value_cool"]
    )

    # def foo3(x):
    #     """
    #     This function is similar to foo2, but it returns electricity values
    #     per unit of cooling for techs that require parasitic electricity demand
    #     """
    #     if "hpl" in x['index']:
    #         return x['parasitic_electricity_demand_fraction']
    #
    #     elif x['parasitic_electricity_demand_fraction'] > 0.0:
    #         return x['parasitic_electricity_demand_fraction'] / x['cooling_fraction']

    # Filter out technologies that requires parasitic electricity
    electr = input_cool[input_cool["parasitic_electricity_demand_fraction"] > 0.0]

    # Make a new column 'value_cool' for calculating values against technologies
    electr["value_cool"] = (
        electr["parasitic_electricity_demand_fraction"] / electr["cooling_fraction"]
    )
    # Filters out technologies requiring saline water supply
    saline_df = input_cool[
        input_cool["technology_name"].str.endswith("ot_saline", na=False)
    ]

    # input_cool_minus_saline_elec_df
    con1 = input_cool["technology_name"].str.endswith("ot_saline", na=False)
    con2 = input_cool["technology_name"].str.endswith("air", na=False)
    icmse_df = input_cool[(~con1) & (~con2)]

    inp = make_df(
        "input",
        node_loc=electr["node_loc"],
        technology=electr["technology_name"],
        year_vtg=electr["year_vtg"],
        year_act=electr["year_act"],
        mode=electr["mode"],
        node_origin=electr["node_origin"],
        commodity="electr",
        level="secondary",
        time="year",
        time_origin="year",
        value=electr["value_cool"],
        unit="GWa",
    )
    # once through and closed loop freshwater
    inp = inp.append(
        make_df(
            "input",
            node_loc=icmse_df["node_loc"],
            technology=icmse_df["technology_name"],
            year_vtg=icmse_df["year_vtg"],
            year_act=icmse_df["year_act"],
            mode=icmse_df["mode"],
            node_origin=icmse_df["node_origin"],
            commodity="freshwater",
            level="water_supply",
            time="year",
            time_origin="year",
            value=icmse_df["value_cool"],
            unit="km3/GWa",
        )
    )
    # saline cooling technologies
    inp = inp.append(
        make_df(
            "input",
            node_loc=saline_df["node_loc"],
            technology=saline_df["technology_name"],
            year_vtg=saline_df["year_vtg"],
            year_act=saline_df["year_act"],
            mode=saline_df["mode"],
            node_origin=saline_df["node_origin"],
            commodity="saline_ppl",
            level="saline_supply",
            time="year",
            time_origin="year",
            value=saline_df["value_cool"],
            unit="km3/GWa",
        )
    )

    # Drops NA values from the value column
    inp = inp.dropna(subset=["value"])

    # append the input data to results
    results["input"] = inp

    # add water consumption as emission factor, also for saline tecs
    emiss_df = input_cool[(~con2)]
    emi = make_df(
        "emission_factor",
        node_loc=emiss_df["node_loc"],
        technology=emiss_df["technology_name"],
        year_vtg=emiss_df["year_vtg"],
        year_act=emiss_df["year_act"],
        mode=emiss_df["mode"],
        emission="fresh_return",
        value=emiss_df["value_return"],
        unit="km3/yr",
    )

    results["emission_factor"] = emi

    # add water return flows for cooling tecs
    # Use share of basin availability to distribute the return flow from
    df_sw = map_basin_region_wat(context)
    df_sw.drop(columns={"mode", "date", "MSGREG"}, inplace=True)
    df_sw.rename(
        columns={"region": "node_dest", "time": "time_dest", "year": "year_act"},
        inplace=True,
    )
    df_sw["time_dest"] = df_sw["time_dest"].astype(str)
    if context.nexus_set == "nexus":
        out = pd.DataFrame()
        for nn in icmse_df.node_loc.unique():
            # input cooling fresh basin
            icfb_df = icmse_df[icmse_df["node_loc"] == nn]
            bs = list(df_node[df_node["region"] == nn]["node"])

            out_t = (
                make_df(
                    "output",
                    node_loc=icfb_df["node_loc"],
                    technology=icfb_df["technology_name"],
                    year_vtg=icfb_df["year_vtg"],
                    year_act=icfb_df["year_act"],
                    mode=icfb_df["mode"],
                    # node_origin=icmse_df["node_origin"],
                    commodity="surfacewater_basin",
                    level="water_avail_basin",
                    time="year",
                    value=icfb_df["value_return"],
                    unit="km3/GWa",
                )
                .pipe(broadcast, node_dest=bs, time_dest=sub_time)
                .merge(df_sw, how="left")
            )
            # multiply by basin water availability share
            out_t["value"] = out_t["value"] * out_t["share"]
            out_t.drop(columns={"share"}, inplace=True)
            out = out.append(out_t)

        out = out.dropna(subset=["value"])
        out.reset_index(drop=True, inplace=True)
        results["output"] = out

    # costs and historical parameters
    path1 = package_data_path("water", "ppl_cooling_tech", FILE1)
    cost = pd.read_csv(path1)
    # Combine technology name to get full cooling tech names
    cost["technology"] = cost["utype"] + "__" + cost["cooling"]
    # Filtering out 2010 data to use for historical values
    input_cool_2010 = input_cool[
        (input_cool["year_act"] == 2010) & (input_cool["year_vtg"] == 2010)
    ]
    # Filter out columns that contain 'mix' in column name
    columns = [col for col in cost.columns if "mix_" in col]
    # Rename column names to R11 to match with the previous df

    cost.rename(columns=lambda name: name.replace("mix_", ""), inplace=True)
    search_cols = [
        col for col in cost.columns if context.regions in col or "technology" in col
    ]
    hold_df = input_cool_2010[
        ["node_loc", "technology_name", "cooling_fraction"]
    ].drop_duplicates()
    search_cols_cooling_fraction = [col for col in search_cols if col != "technology"]

    def shares(x, context):
        """Process share and cooling fraction.
        Returns
        -------
        Product of value of shares of cooling technology types of regions with
        corresponding cooling fraction
        """
        for col in search_cols_cooling_fraction:
            # MAPPING ISOCODE to region name, assume one country only
            col2 = context.map_ISO_c[col] if context.type_reg == "country" else col
            cooling_fraction = hold_df[
                (hold_df["node_loc"] == col2)
                & (hold_df["technology_name"] == x["technology"])
            ]["cooling_fraction"]
            x[col] = x[col] * cooling_fraction

        results = []
        for i in x:
            if isinstance(i, str):
                results.append(i)
            else:
                if not len(i):
                    return pd.Series(
                        [i for i in range(len(search_cols) - 1)] + ["delme"],
                        index=search_cols,
                    )
                else:
                    results.append(float(i))
        return pd.Series(results, index=search_cols)

    # Apply function to the
    hold_cost = cost[search_cols].apply(shares, axis=1, context=context)
    hold_cost = hold_cost[hold_cost["technology"] != "delme"]

    def hist_act(x, context):
        """Calculate historical activity of cooling technology.
        The data for shares is read from ``cooltech_cost_and_shares_ssp_msg.csv``
        Returns
        -------
        hist_activity(cooling_tech) = hist_activitiy(parent_technology) * share
        *cooling_fraction
        """
        tech_df = hold_cost[
            hold_cost["technology"].str.startswith(x.technology)
        ]  # [x.node_loc]

        node_search = (
            context.regions if context.type_reg == "country" else x["node_loc"]
        )

        node_loc = x["node_loc"]
        technology = x["technology"]
        cooling_technologies = list(tech_df["technology"])
        new_values = tech_df[node_search] * x.value

        return [
            [
                node_loc,
                technology,
                cooling_technology,
                x.year_act,
                x.value,
                new_value,
                x.unit,
            ]
            for new_value, cooling_technology in zip(new_values, cooling_technologies)
        ]

    changed_value_series = ref_hist_act.apply(hist_act, axis=1, context=context)
    changed_value_series_flat = [
        row for series in changed_value_series for row in series
    ]
    columns = [
        "node_loc",
        "technology",
        "cooling_technology",
        "year_act",
        "value",
        "new_value",
        "unit",
    ]
    # dataframe for historical activities of cooling techs
    act_value_df = pd.DataFrame(changed_value_series_flat, columns=columns)

    def hist_cap(x, context):
        """Calculate historical capacity of cooling technology.
        The data for shares is read from ``cooltech_cost_and_shares_ssp_msg.csv``
        Returns
        -------
        hist_new_capacity(cooling_tech) = historical_new_capacity(parent_technology)*
        share * cooling_fraction
        """
        tech_df = hold_cost[
            hold_cost["technology"].str.startswith(x.technology)
        ]  # [x.node_loc]
        if context.type_reg == "country":
            node_search = context.regions
        else:
            node_search = x["node_loc"]  # R11_EEU
        node_loc = x["node_loc"]
        technology = x["technology"]
        cooling_technologies = list(tech_df["technology"])
        new_values = tech_df[node_search] * x.value

        return [
            [
                node_loc,
                technology,
                cooling_technology,
                x.year_vtg,
                x.value,
                new_value,
                x.unit,
            ]
            for new_value, cooling_technology in zip(new_values, cooling_technologies)
        ]

    changed_value_series = ref_hist_cap.apply(hist_cap, axis=1, context=context)
    changed_value_series_flat = [
        row for series in changed_value_series for row in series
    ]
    columns = [
        "node_loc",
        "technology",
        "cooling_technology",
        "year_vtg",
        "value",
        "new_value",
        "unit",
    ]
    cap_value_df = pd.DataFrame(changed_value_series_flat, columns=columns)

    # Make model compatible df for historical activitiy
    h_act = make_df(
        "historical_activity",
        node_loc=act_value_df["node_loc"],
        technology=act_value_df["cooling_technology"],
        year_act=act_value_df["year_act"],
        mode="M1",
        time="year",
        value=act_value_df["new_value"],
        # TODO finalize units
        unit="GWa",
    )

    results["historical_activity"] = h_act
    # Make model compatible df for histroical new capacity
    h_cap = make_df(
        "historical_new_capacity",
        node_loc=cap_value_df["node_loc"],
        technology=cap_value_df["cooling_technology"],
        year_vtg=cap_value_df["year_vtg"],
        value=cap_value_df["new_value"],
        unit="GWa",
    )

    results["historical_new_capacity"] = h_cap

    # Add upper bound for seawater cooling
    # sums up all the historical activities of seawater cooling technologies
    # h_act_saline = h_act[h_act["technology"].str.endswith("saline")]
    # h_act_saline = h_act_saline[h_act_saline["year_act"] == 2015]
    # h_act_saline.drop(columns=["year_act", "mode", "time", "unit"], inplace=True)
    # h_act_saline = h_act_saline.groupby(["node_loc"]).sum()

    # inp_saline = inp[inp["technology"].str.endswith("ot_saline")]
    # inp_saline = inp_saline[
    #     (inp_saline["year_vtg"] == 2015) & (inp_saline["year_act"] == 2015)
    # ]
    # inp_saline.drop(
    #     columns=[
    #         "year_vtg",
    #         "commodity",
    #         "year_act",
    #         "mode",
    #         "level",
    #         "time",
    #         "time_origin",
    #         "unit",
    #         "node_origin",
    #     ],
    #     inplace=True,
    # )
    # water_fr = inp_saline.groupby(["node_loc"]).mean()
    # # multiplying input values of water withdrawal with
    # bound_saline = water_fr.mul(h_act_saline)

    # bound_up = make_df(
    #     "bound_activity_up",
    #     node_loc=bound_saline.index,
    #     technology="extract_salinewater",
    #     mode="M1",
    #     time="year",
    #     value=bound_saline["value"].values,
    #     unit="km3/year",
    # ).pipe(broadcast, year_act=info.Y)

    # results["bound_activity_up"] = bound_up

    # Filter out just cl_fresh & air technologies for adding inv_cost in model,
    # The rest of technologies are assumed to have costs included in parent technologies
    # con3 = cost['technology'].str.endswith("cl_fresh")
    # con4 = cost['technology'].str.endswith("air")
    # con5 = cost.technology.isin(input_cool['technology_name'])
    # inv_cost = cost[(con3) | (con4)]
    inv_cost = cost.copy()
    # Manually removing extra technologies not required
    # TODO make it automatic to not include the names manually
    techs_to_remove = [
        "mw_ppl__ot_fresh",
        "mw_ppl__ot_saline",
        "mw_ppl__cl_fresh",
        "mw_ppl__air",
        "nuc_fbr__ot_fresh",
        "nuc_fbr__ot_saline",
        "nuc_fbr__cl_fresh",
        "nuc_fbr__air",
        "nuc_htemp__ot_fresh",
        "nuc_htemp__ot_saline",
        "nuc_htemp__cl_fresh",
        "nuc_htemp__air",
    ]
    inv_cost = inv_cost[~inv_cost["technology"].isin(techs_to_remove)]
    # Converting the cost to USD/GW
    inv_cost["investment_USD_per_GW_mid"] = (
        inv_cost["investment_million_USD_per_MW_mid"] * 1e3
    )

    inv_cost = (
        make_df(
            "inv_cost",
            technology=inv_cost["technology"],
            value=inv_cost["investment_USD_per_GW_mid"],
            unit="USD/GWa",
        )
        .pipe(same_node)
        .pipe(broadcast, node_loc=node_region, year_vtg=info.Y)
    )

    results["inv_cost"] = inv_cost

    # Addon conversion
    adon_df = input_cool.copy()
    # Add 'cooling_' before name of parent technologies that are type_addon
    # nomenclature
    adon_df["tech"] = "cooling__" + adon_df["index"].astype(str)
    # technology : 'parent technology' and type_addon is type of addons such
    # as 'cooling__bio_hpl'
    addon_df = make_df(
        "addon_conversion",
        node=adon_df["node_loc"],
        technology=adon_df["index"],
        year_vtg=adon_df["year_vtg"],
        year_act=adon_df["year_act"],
        mode=adon_df["mode"],
        time="year",
        type_addon=adon_df["tech"],
        value=adon_df["cooling_fraction"],
        unit="km3/GWa",
    )

    results["addon_conversion"] = addon_df

    # Addon_lo will remain 1 for all cooling techs so it allows 100% activity of
    # parent technologies
    addon_lo = make_matched_dfs(addon_df, addon_lo=1)
    results["addon_lo"] = addon_lo["addon_lo"]

    # technical lifetime
    # make_matched_dfs didn't map all technologies
    # tl = make_matched_dfs(inv_cost,
    #                       technical_lifetime = 30)
    year = info.Y
    if 2010 in year:
        pass
    else:
        year.insert(0, 2010)

    tl = (
        make_df(
            "technical_lifetime",
            technology=inp["technology"].drop_duplicates(),
            value=30,
            unit="year",
        )
        .pipe(broadcast, year_vtg=year, node_loc=node_region)
        .pipe(same_node)
    )

    results["technical_lifetime"] = tl

    cap_fact = make_matched_dfs(inp, capacity_factor=1)
    # Climate Impacts on freshwater cooling capacity
    # Taken from
    # https://www.sciencedirect.com/science/article/
    #  pii/S0959378016301236?via%3Dihub#sec0080
    if context.RCP == "no_climate":
        df = cap_fact["capacity_factor"]
    else:
        df = cap_fact["capacity_factor"]
        # reading ppl cooling impact dataframe
        path = package_data_path(
            "water", "ppl_cooling_tech", "power_plant_cooling_impact_MESSAGE.xlsx"
        )
        df_impact = pd.read_excel(path, sheet_name=f"{context.regions}_{context.RCP}")

        for n in df_impact["node"]:
            conditions = [
                df["technology"].str.contains("fresh")
                & (df["year_act"] >= 2025)
                & (df["year_act"] < 2050)
                & (df["node_loc"] == n),
                df["technology"].str.contains("fresh")
                & (df["year_act"] >= 2050)
                & (df["year_act"] < 2070)
                & (df["node_loc"] == n),
                df["technology"].str.contains("fresh")
                & (df["year_act"] >= 2070)
                & (df["node_loc"] == n),
            ]

            choices = [
                df_impact[(df_impact["node"] == n)]["2025s"],
                df_impact[(df_impact["node"] == n)]["2050s"],
                df_impact[(df_impact["node"] == n)]["2070s"],
            ]

            df["value"] = np.select(conditions, choices, default=df["value"])

    results["capacity_factor"] = df
    # results = {par_name: pd.concat(dfs) for par_name, dfs in results.items()}

    # growth activity low to allow the cooling techs to be operational
    g_lo = make_df(
        "growth_activity_lo",
        technology=inp["technology"].drop_duplicates(),
        value=-0.05,
        unit="%",
        time="year",
    ).pipe(broadcast, year_act=info.Y, node_loc=node_region)
    # Alligining certain technologies with growth constriants
    g_lo.loc[g_lo["technology"].str.contains("bio_ppl|loil_ppl"), "value"] = -0.5
    g_lo.loc[g_lo["technology"].str.contains("coal_ppl_u|coal_ppl"), "value"] = -0.5
    g_lo.loc[
        (g_lo["technology"].str.contains("coal_ppl_u|coal_ppl"))
        & (g_lo["node_loc"].str.contains("CPA|PAS")),
        "value",
    ] = -1
    results["growth_activity_lo"] = g_lo

    # growth activity up on saline water
    inp_saline = inp[inp["technology"].str.endswith("ot_saline")]

    g_up = make_df(
        "growth_activity_up",
        technology=inp_saline["technology"].drop_duplicates(),
        value=0.05,
        unit="%",
        time="year",
    ).pipe(broadcast, year_act=info.Y, node_loc=node_region)
    results["growth_activity_up"] = g_up

    # # adding initial activity
    # in_lo = h_act.copy()
    # in_lo.drop(columns='mode', inplace=True)
    # in_lo = in_lo[in_lo['year_act'] == 2015]
    # in_lo_1 = make_df('initial_activity_lo',
    #                   node_loc=in_lo['node_loc'],
    #                   technology=in_lo['technology'],
    #                   time='year',
    #                   value=in_lo['value'],
    #                   unit='GWa').pipe(broadcast, year_act=[2015, 2020])
    # results['initial_activity_lo'] = in_lo_1

    return results


# Water use & electricity for non-cooling technologies
def non_cooling_tec(context):
    """Process data for water usage of power plants (non-cooling technology related).
    Water withdrawal values for power plants are read in from
    ``tech_water_performance_ssp_msg.csv``
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'.
        Values are data frames ready for :meth:`~.Scenario.add_par`.
        Years in the data include the model horizon indicated by
        ``context["transport build info"]``, plus the additional year 2010.
    """
    results = {}

    FILE = "tech_water_performance_ssp_msg.csv"
    path = package_data_path("water", "ppl_cooling_tech", FILE)
    df = pd.read_csv(path)
    cooling_df = df.loc[df["technology_group"] == "cooling"]
    # Separate a column for parent technologies of respective cooling
    # techs
    cooling_df["parent_tech"] = (
        cooling_df["technology_name"]
        .apply(lambda x: pd.Series(str(x).split("__")))
        .drop(columns=1)
    )
    non_cool_df = df[
        (df["technology_group"] != "cooling")
        & (df["water_supply_type"] == "freshwater_supply")
    ]

    scen = context.get_scenario()
    tec_lt = scen.par("technical_lifetime")
    all_tech = list(tec_lt["technology"].unique())
    # all_tech = list(scen.set("technology"))
    tech_non_cool_csv = list(non_cool_df["technology_name"])
    techs_to_remove = [tec for tec in tech_non_cool_csv if tec not in all_tech]

    non_cool_df = non_cool_df[~non_cool_df["technology_name"].isin(techs_to_remove)]
    non_cool_df = non_cool_df.rename(columns={"technology_name": "technology"})

    non_cool_df["value"] = (
        non_cool_df["water_withdrawal_mid_m3_per_output"] * 60 * 60 * 24 * 365 * (1e-9)
    )

    non_cool_tech = list(non_cool_df["technology"].unique())

    n_cool_df = scen.par("output", {"technology": non_cool_tech})
    n_cool_df = n_cool_df[
        (n_cool_df["node_loc"] != "R11_GLB") & (n_cool_df["node_dest"] != "R11_GLB")
    ]
    n_cool_df_merge = pd.merge(n_cool_df, non_cool_df, on="technology", how="right")
    n_cool_df_merge.dropna(inplace=True)

    # Input dataframe for non cooling technologies
    # only water withdrawals are being taken
    # Only freshwater supply is assumed for simplicity
    inp_n_cool = make_df(
        "input",
        technology=n_cool_df_merge["technology"],
        value=n_cool_df_merge["value_y"],
        unit="km3/GWa",
        level="water_supply",
        commodity="freshwater",
        time_origin="year",
        mode="M1",
        time="year",
        year_vtg=n_cool_df_merge["year_vtg"].astype(int),
        year_act=n_cool_df_merge["year_act"].astype(int),
        node_loc=n_cool_df_merge["node_loc"],
        node_origin=n_cool_df_merge["node_dest"],
    )

    # append the input data to results
    results["input"] = inp_n_cool

    return results



import logging
from abc import abstractmethod
from pathlib import Path
from typing import Any, Dict, Mapping, Optional, Sequence, Tuple, cast

import pandas as pd
from genno import Quantity
from genno.operator import concat

log = logging.getLogger(__name__)

try:
    import message_data
except ImportError:
    log.warning("message_data is not installed or cannot be imported")
    MESSAGE_DATA_PATH: Optional[Path] = None
    HAS_MESSAGE_DATA = False
else:  # pragma: no cover  (needs message_data)
    # Root directory of the message_data repository.
    MESSAGE_DATA_PATH = Path(message_data.__file__).parents[1]
    HAS_MESSAGE_DATA = True

# Directory containing message_ix_models.__init__
MESSAGE_MODELS_PATH = Path(__file__).parents[1]

#: Package data already loaded with :func:`load_package_data`.
PACKAGE_DATA: Dict[str, Any] = dict()

#: Data already loaded with :func:`load_private_data`.
PRIVATE_DATA: Dict[str, Any] = dict()


__all__ = [
    "HAS_MESSAGE_DATA",
    "Adapter",
    "MappingAdapter",
]


class Adapter:
    """Adapt `data`.

    Adapter is an abstract base class for tools that adapt data in any way, e.g.
    between different code lists for certain dimensions. An instance of an Adapter can
    be called with any of the following as `data`:

    - :class:`genno.Quantity`,
    - :class:`pandas.DataFrame`, or
    - :class:`dict` mapping :class:`str` parameter names to values (either of the above
      types).

    …and will return data of the same type.

    Subclasses can implement different adapter logic by overriding the abstract
    :meth:`adapt` method.
    """

    def __call__(self, data):
        if isinstance(data, Quantity):
            return self.adapt(data)
        elif isinstance(data, pd.DataFrame):
            # Convert to Quantity
            idx_cols = list(filter(lambda c: c not in ("value", "unit"), data.columns))
            qty = Quantity.from_series(data.set_index(idx_cols)["value"])

            # Store units
            if "unit" in data.columns:
                units = data["unit"].unique()
                assert 1 == len(units), f"Non-unique units {units}"
                unit = units[0]
            else:
                unit = ""  # dimensionless

            # Adapt, convert back to pd.DataFrame, return
            return self.adapt(qty).to_dataframe().assign(unit=unit).reset_index()
        elif isinstance(data, Mapping):
            return {par: self(value) for par, value in data.items()}
        else:
            raise TypeError(type(data))

    @abstractmethod
    def adapt(self, qty: Quantity) -> Quantity:
        """Adapt data."""


class MappingAdapter(Adapter):
    """Adapt data using mappings for 1 or more dimension(s).

    Parameters
    ----------
    maps : dict of sequence of (str, str)
        Keys are names of dimensions. Values are sequences of 2-tuples; each tuple
        consists of an original label and a target label.

    Examples
    --------
    >>> a = MappingAdapter({"foo": [("a", "x"), ("a", "y"), ("b", "z")]})
    >>> df = pd.DataFrame(
    ...     [["a", "m", 1], ["b", "n", 2]], columns=["foo", "bar", "value"]
    ... )
    >>> a(df)
      foo  bar  value
    0   x    m      1
    1   y    m      1
    2   z    n      2
    """

    maps: Mapping

    def __init__(self, maps: Mapping[str, Sequence[Tuple[str, str]]]):
        self.maps = maps

    def adapt(self, qty: Quantity) -> Quantity:
        result = qty

        for dim, labels in self.maps.items():
            if dim not in qty.dims:  # type: ignore [attr-defined]
                continue
            result = concat(
                *[
                    qty.sel({dim: label[0]}, drop=True).expand_dims({dim: [label[1]]})
                    for label in labels
                ]
            )

        return result


def _load(
    var: Dict, base_path: Path, *parts: str, default_suffix: Optional[str] = None
) -> Any:
    """Helper for :func:`.load_package_data` and :func:`.load_private_data`."""
    key = " ".join(parts)
    if key in var:
        log.debug(f"{repr(key)} already loaded; skip")
        return var[key]

    path = _make_path(base_path, *parts, default_suffix=default_suffix)

    if path.suffix == ".yaml":
        import yaml

        with open(path, encoding="utf-8") as f:
            var[key] = yaml.safe_load(f)
    else:
        raise ValueError(path.suffix)

    return var[key]


def _make_path(
    base_path: Path, *parts: str, default_suffix: Optional[str] = None
) -> Path:
    p = base_path.joinpath(*parts)
    return p.with_suffix(p.suffix or default_suffix) if default_suffix else p


def load_package_data(*parts: str, suffix: Optional[str] = ".yaml") -> Any:
    """Load a :mod:`message_ix_models` package data file and return its contents.

    Data is re-used if already loaded.

    Example
    -------

    The single call:

    >>> info = load_package_data("node", "R11")

    1. loads the metadata file :file:`data/node/R11.yaml`, parsing its contents,
    2. stores those values at ``PACKAGE_DATA["node R11"]`` for use by other code, and
    3. returns the loaded values.

    Parameters
    ----------
    parts : iterable of str
        Used to construct a path under :file:`message_ix_models/data/`.
    suffix : str, *optional*
        File name suffix, including, the ".", e.g. :file:`.yaml`.

    Returns
    -------
    dict
        Configuration values that were loaded.
    """
    return _load(
        PACKAGE_DATA,
        MESSAGE_MODELS_PATH / "data",
        *parts,
        default_suffix=suffix,
    )


def load_private_data(*parts: str) -> Mapping:  # pragma: no cover (needs message_data)
    """Load a private data file from :mod:`message_data` and return its contents.

    Analogous to :func:`load_package_data`, but for non-public data.

    Parameters
    ----------
    parts : iterable of str
        Used to construct a path under :file:`data/` in the :mod:`message_data`
        repository.

    Returns
    -------
    dict
        Configuration values that were loaded.

    Raises
    ------
    RuntimeError
        if :mod:`message_data` is not installed.
    """
    if MESSAGE_DATA_PATH is None:
        raise RuntimeError("message_data is not installed")

    return _load(PRIVATE_DATA, MESSAGE_DATA_PATH / "data", *parts)


def local_data_path(*parts) -> Path:
    """Construct a path for local data.

    The setting ``message local data`` in the user's :ref:`ixmp configuration file
    <ixmp:configuration>` is used as a base path. If this is not configured, the
    current working directory is used.

    Parameters
    ----------
    parts : sequence of str or Path
        Joined to the base path using :meth:`.Path.joinpath`.

    See also
    --------
    :ref:`Choose locations for data <local-data>`
    """
    import ixmp

    # The default value, Path.cwd(), is set in context.py
    return _make_path(Path(ixmp.config.get("message local data")), *parts)


def package_data_path(*parts) -> Path:
    """Construct a path to a file under :file:`message_ix_models/data/`.

    Use this function to access data packaged and installed with
    :mod:`message_ix_models`.

    Parameters
    ----------
    parts : sequence of str or Path
        Joined to the base path using :meth:`~pathlib.PurePath.joinpath`.

    See also
    --------
    :ref:`Choose locations for data <package-data>`
    """
    return _make_path(MESSAGE_MODELS_PATH / "data", *parts)


def private_data_path(*parts) -> Path:  # pragma: no cover (needs message_data)
    """Construct a path to a file under :file:`data/` in :mod:`message_data`.

    Use this function to access non-public (e.g. embargoed or proprietary) data stored
    in the :mod:`message_data` repository.

    Parameters
    ----------
    parts : sequence of str or Path
        Joined to the base path using :meth:`~pathlib.PurePath.joinpath`.

    See also
    --------
    :ref:`Choose locations for data <private-data>`
    """
    return _make_path(cast(Path, MESSAGE_DATA_PATH) / "data", *parts)



import logging
from copy import copy
from functools import partial
from urllib.parse import urlunsplit

import message_ix
from sdmx.model.v21 import Code

import message_ix_models
from message_ix_models import ScenarioInfo, Spec
from message_ix_models.util import eval_anno

from .build import apply_spec
from .config import Config
from .data import get_data
from .structure import codelists, get_codes

log = logging.getLogger(__name__)


def _default_first(kind, default):
    return [default] + list(filter(lambda id: id != default, codelists(kind)))


#: Deprecated; use :class:`.model.Config` instead.
SETTINGS = dict(
    # Place the default value first
    regions=_default_first("node", "R14"),
    years=_default_first("year", "B"),
    res_with_dummies=[False, True],
)


def create_res(context, quiet=True):
    """Create a 'bare' MESSAGEix-GLOBIOM reference energy system (RES).

    Parameters
    ----------
    context : Context
        :attr:`.Context.scenario_info` determines the model name and scenario name of
        the created Scenario. If not provided, the defaults are:

        - Model name generated by :func:`name`.
        - Scenario name "baseline".
    quiet : bool, *optional*
        Passed to `quiet` argument of :func:`.build.apply_spec`.

    Returns
    -------
    message_ix.Scenario
        A scenario as described by :func:`.bare.get_spec`, prepared using
        :func:`.apply_spec`.
    """
    mp = context.get_platform()

    # Retrieve the spec; this also sets defaults expected by name()
    spec = get_spec(context)

    # Model and scenario name for the RES
    args = dict(
        mp=mp,
        model=context.scenario_info.get("model", name(context)),
        scenario=context.scenario_info.get("scenario", "baseline"),
        version="new",
    )

    # TODO move this to ixmp as a method similar to ixmp.util.parse_url()
    url = urlunsplit(
        ("ixmp", mp.name, args["model"] + "/" + args["scenario"], "", args["version"])
    )
    log.info(f"Create {repr(url)}")

    # Create the Scenario
    scenario = message_ix.Scenario(**args)

    # TODO move to message_ix
    scenario.init_par("MERtoPPP", ["node", "year"])

    # Apply the spec
    apply_spec(
        scenario,
        spec,
        data=partial(get_data, context=context, spec=spec),
        quiet=quiet,
        message=f"Create using message-ix-models {message_ix_models.__version__}",
    )

    return scenario


def get_spec(context) -> Spec:
    """Return the spec for the MESSAGE-GLOBIOM global model RES.

    If :attr:`.Config.res_with_dummies` is set, additional elements are added:

    - ``commodity``: "dummy"
    - ``technology``: "dummy", "dummy supply"

    These **may** be used for testing purposes, but **should not** be used in production
    models.

    Returns
    -------
    :class:`dict` of :class:`.ScenarioInfo` objects
    """
    context.setdefault("model", Config())

    add = ScenarioInfo()

    # Add technologies
    add.set["technology"] = copy(get_codes("technology"))

    # Add regions

    # Load configuration for the specified region mapping
    nodes = get_codes(f"node/{context.model.regions}")

    # Top-level "World" node
    world = nodes[nodes.index("World")]

    # Set elements: World, followed by the direct children of World
    add.set["node"] = [world] + world.child

    # Add relations
    add.set["relation"] = get_codes(f"relation/{context.model.relations}")

    # Initialize time periods
    add.year_from_codes(get_codes(f"year/{context.model.years}"))

    # Add levels
    add.set["level"] = get_codes("level")

    # Add commodities
    add.set["commodity"] = get_codes("commodity")

    # Add units, associated with commodities
    units = set(eval_anno(commodity, "unit") for commodity in add.set["commodity"])
    # Deduplicate by converting to a set and then back; not strictly necessary,
    # but reduces duplicate log entries
    add.set["unit"] = sorted(filter(None, units))

    if context.model.res_with_dummies:
        # Add dummy technologies
        add.set["technology"].extend([Code(id="dummy"), Code(id="dummy source")])
        # Add a dummy commodity
        add.set["commodity"].append(Code(id="dummy"))

    # The RES is the base, so does not require or remove any elements
    return Spec(add=add)


def name(context):
    """Generate a candidate name for a model given `context`.

    The name has a form like::

        MESSAGEix-GLOBIOM R99 YA +D

    where:

    - "R99" is the node list/regional aggregation.
    - "YA" indicates the year codelist (:doc:`/pkg-data/year`).
    - "+D" appears if :attr:`.Config.res_with_dummies` is true.

    """
    cfg = context.model
    return f"MESSAGEix-GLOBIOM {cfg.regions} Y{cfg.years}" + (
        " +D" if cfg.res_with_dummies else ""
    )



"""Prepare data for adding demands"""

import os

import numpy as np
import pandas as pd
import xarray as xr
from message_ix import make_df

from message_ix_models.util import broadcast, package_data_path


def get_basin_sizes(basin, node):
    """Returns the sizes of developing and developed basins for a given node"""
    temp = basin[basin["BCU_name"] == node]
    sizes = temp.pivot_table(index=["STATUS"], aggfunc="size")
    return sizes["DEV"], sizes["IND"]


def set_target_rate(df, node, year, target):
    """Sets the target value for a given node and year"""
    indices = df[df["node"] == node][df[df["node"] == node]["year"] == year].index
    for index in indices:
        if (
            df[df["node"] == node][df[df["node"] == node]["year"] == year].at[
                index, "value"
            ]
            < target
        ):
            df.at[index, "value"] = target


def set_target_rate_developed(df, node, target):
    """Sets target rate for a developed basin"""
    set_target_rate(df, node, 2030, target)


def set_target_rate_developing(df, node, target):
    """Sets target rate for a developing basin"""
    for i in df.index:
        if df.at[i, "node"] == node and df.at[i, "year"] == 2030:
            value_2030 = df.at[i, "value"]
            break

    set_target_rate(
        df,
        node,
        2035,
        (value_2030 + target) / 2,
    )
    set_target_rate(df, node, 2040, target)


def set_target_rates(df, basin, val):
    """Sets target rates for all nodes in a given basin"""
    for node in df.node.unique():
        dev_size, ind_size = get_basin_sizes(basin, node)
        if dev_size >= ind_size:
            set_target_rate_developed(df, node, val)
        else:
            set_target_rate_developing(df, node, val)


def target_rate(df, basin, val):
    """
    Sets target connection and sanitation rates for SDG scenario.
    The function filters out the basins as developing and
    developed based on the countries overlapping basins.
    If the number of developing countries in the basins are
    more than basin is categorized as developing and vice versa.
    If the number of developing and developed countries are equal
    in a basin, then the basin is assumed developing.
    For developed basins, target is set at 2030.
    For developing basins, the access target is set at
    2040 and 2035 target is the average of
    2030 original rate and 2040 target.
    Returns:
        df (pandas.DataFrame): Data frame with updated value column.
    """
    set_target_rates(df, basin, val)
    return df


def target_rate_trt(df, basin):
    """
    Sets target treatment rates for SDG scenario. The target value for
    developed and developing region is making sure that the amount of untreated
    wastewater is halved beyond 2030 & 2040 respectively.
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
    """

    value = []
    for i in df.node.unique():
        temp = basin[basin["BCU_name"] == i]

        sizes = temp.pivot_table(index=["STATUS"], aggfunc="size")

        if len(sizes) > 1:
            if sizes["DEV"] > sizes["IND"] or sizes["DEV"] == sizes["IND"]:
                for j in df[df["node"] == i][df[df["node"] == i]["year"] >= 2040].index:
                    temp = df[df["node"] == i][df[df["node"] == i]["year"] >= 2040].at[
                        j, "value"
                    ]
                    temp = temp + (1 - temp) / 2
                    value.append([j, np.float64(temp)])
            else:
                for j in df[df["node"] == i][df[df["node"] == i]["year"] >= 2030].index:
                    temp = df[df["node"] == i][df[df["node"] == i]["year"] >= 2030].at[
                        j, "value"
                    ]
                    temp = temp + (1 - temp) / 2
                    value.append([j, np.float64(temp)])
        else:
            if sizes.index[0] == "DEV":
                for j in df[df["node"] == i][df[df["node"] == i]["year"] >= 2040].index:
                    temp = df[df["node"] == i][df[df["node"] == i]["year"] >= 2040].at[
                        j, "value"
                    ]
                    temp = temp + (1 - temp) / 2
                    value.append([j, np.float64(temp)])
            else:
                for j in df[df["node"] == i][df[df["node"] == i]["year"] >= 2030].index:
                    temp = df[df["node"] == i][df[df["node"] == i]["year"] >= 2030].at[
                        j, "value"
                    ]
                    temp = temp + (1 - temp) / 2
                    value.append([j, np.float64(temp)])

    valuetest = pd.DataFrame(data=value, columns=["Index", "Value"])

    for i in range(len(valuetest["Index"])):
        df.at[valuetest["Index"][i], "Value"] = valuetest["Value"][i]

    real_value = df["Value"].combine_first(df["value"])

    df.drop(["value", "Value"], axis=1, inplace=True)

    df["value"] = real_value
    return df


def add_sectoral_demands(context):
    """
    Adds water sectoral demands
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'. Values
        are data frames ready for :meth:`~.Scenario.add_par`.
    """

    # define an empty dictionary
    results = {}

    # Reference to the water configuration
    info = context["water build info"]

    # defines path to read in demand data
    region = f"{context.regions}"
    sub_time = context.time
    path = package_data_path("water", "demands", "harmonized", region, ".")
    # make sure all of the csvs have format, otherwise it might not work
    list_of_csvs = list(path.glob("ssp2_regional_*.csv"))
    # define names for variables
    fns = [os.path.splitext(os.path.basename(x))[0] for x in list_of_csvs]
    fns = " ".join(fns).replace("ssp2_regional_", "").split()
    # dictionary for reading csv files
    d = {}

    for i in range(len(fns)):
        d[fns[i]] = pd.read_csv(list_of_csvs[i])

    # d is a dictionary that have ist of dataframes read in this folder
    dfs = {}
    for key, df in d.items():
        df.rename(columns={"Unnamed: 0": "year"}, inplace=True)
        df.index = df["year"]
        df = df.drop(columns=["year"])
        dfs[key] = df

    # convert the dictionary of dataframes to xarray
    df_x = xr.Dataset(dfs).to_array()
    df_x_interp = df_x.interp(year=[2015, 2025, 2035, 2045, 2055])
    df_x_c = df_x.combine_first(df_x_interp)
    # Unstack xarray back to pandas dataframe
    df_f = df_x_c.to_dataframe("").unstack()

    # Format the dataframe to be compatible with message format
    df_dmds = df_f.stack().reset_index(level=0).reset_index()
    df_dmds.columns = ["year", "node", "variable", "value"]
    df_dmds.sort_values(["year", "node", "variable", "value"], inplace=True)

    df_dmds["time"] = "year"

    # Write final interpolated values as csv
    # df2_f.to_csv('final_interpolated_values.csv')

    # if we are using sub-annual timesteps we replace the rural and municipal
    # withdrawals and return flows with monthly data and also add industrial
    if "year" not in context.time:
        PATH = package_data_path(
            "water", "demands", "harmonized", region, "ssp2_m_water_demands.csv"
        )
        df_m = pd.read_csv(PATH)
        df_m.value *= 30  # from mcm/day to mcm/month
        df_m["sector"][df_m["sector"] == "industry"] = "manufacturing"
        df_m["variable"] = df_m["sector"] + "_" + df_m["type"] + "_baseline"
        df_m["variable"].replace(
            "urban_withdrawal_baseline", "urban_withdrawal2_baseline", inplace=True
        )
        df_m["variable"].replace(
            "urban_return_baseline", "urban_return2_baseline", inplace=True
        )
        df_m = df_m[["year", "pid", "variable", "value", "month"]]
        df_m.columns = ["year", "node", "variable", "value", "time"]

        # remove yearly parts from df_dms
        df_dmds = df_dmds[
            ~df_dmds["variable"].isin(
                [
                    "urban_withdrawal2_baseline",
                    "rural_withdrawal_baseline",
                    "manufacturing_withdrawal_baseline",
                    "manufacturing_return_baseline",
                    "urban_return2_baseline",
                    "rural_return_baseline",
                ]
            )
        ]
        # attach the monthly demand
        df_dmds = pd.concat([df_dmds, df_m])

    urban_withdrawal_df = df_dmds[df_dmds["variable"] == "urban_withdrawal2_baseline"]
    rual_withdrawal_df = df_dmds[df_dmds["variable"] == "rural_withdrawal_baseline"]
    industrial_withdrawals_df = df_dmds[
        df_dmds["variable"] == "manufacturing_withdrawal_baseline"
    ]
    industrial_return_df = df_dmds[
        df_dmds["variable"] == "manufacturing_return_baseline"
    ]
    urban_return_df = df_dmds[df_dmds["variable"] == "urban_return2_baseline"]
    urban_return_df.reset_index(drop=True, inplace=True)
    rural_return_df = df_dmds[df_dmds["variable"] == "rural_return_baseline"]
    rural_return_df.reset_index(drop=True, inplace=True)
    urban_connection_rate_df = df_dmds[
        df_dmds["variable"] == "urban_connection_rate_baseline"
    ]
    urban_connection_rate_df.reset_index(drop=True, inplace=True)
    rural_connection_rate_df = df_dmds[
        df_dmds["variable"] == "rural_connection_rate_baseline"
    ]
    rural_connection_rate_df.reset_index(drop=True, inplace=True)

    urban_treatment_rate_df = df_dmds[
        df_dmds["variable"] == "urban_treatment_rate_baseline"
    ]
    urban_treatment_rate_df.reset_index(drop=True, inplace=True)

    rural_treatment_rate_df = df_dmds[
        df_dmds["variable"] == "rural_treatment_rate_baseline"
    ]
    rural_treatment_rate_df.reset_index(drop=True, inplace=True)

    df_recycling = df_dmds[df_dmds["variable"] == "urban_recycling_rate_baseline"]
    df_recycling.reset_index(drop=True, inplace=True)

    all_rates_base = pd.concat(
        [
            urban_connection_rate_df,
            rural_connection_rate_df,
            urban_treatment_rate_df,
            rural_treatment_rate_df,
            df_recycling,
        ]
    )

    if context.SDG != "baseline":
        # only if SDG exactly equal to SDG, otherwise other policies are possible
        if context.SDG == "SDG":
            # reading basin mapping to countries
            FILE2 = f"basins_country_{context.regions}.csv"
            PATH = package_data_path("water", "delineation", FILE2)

            df_basin = pd.read_csv(PATH)

            # Applying 80% sanitation rate for rural sanitation
            rural_treatment_rate_df = rural_treatment_rate_df_sdg = target_rate(
                rural_treatment_rate_df, df_basin, 0.8
            )
            # Applying 95% sanitation rate for urban sanitation
            urban_treatment_rate_df = urban_treatment_rate_df_sdg = target_rate(
                urban_treatment_rate_df, df_basin, 0.95
            )
            # Applying 99% connection rate for urban infrastructure
            urban_connection_rate_df = urban_connection_rate_df_sdg = target_rate(
                urban_connection_rate_df, df_basin, 0.99
            )
            # Applying 80% connection rate for rural infrastructure
            rural_connection_rate_df = rural_connection_rate_df_sdg = target_rate(
                rural_connection_rate_df, df_basin, 0.8
            )
            # Applying sdg6 waste water treatment target
            df_recycling = df_recycling_sdg = target_rate_trt(df_recycling, df_basin)

        else:
            pol_scen = context.SDG

            # check if data is there
            check_dm = df_dmds[
                df_dmds["variable"] == "urban_connection_rate_" + pol_scen
            ]
            if check_dm.empty:
                raise ValueError(f"Policy data is missing for the {pol_scen} scenario.")
            urban_connection_rate_df = urban_connection_rate_df_sdg = df_dmds[
                df_dmds["variable"] == "urban_connection_rate_" + pol_scen
            ]
            urban_connection_rate_df.reset_index(drop=True, inplace=True)
            rural_connection_rate_df = rural_connection_rate_df_sdg = df_dmds[
                df_dmds["variable"] == "rural_connection_rate_" + pol_scen
            ]
            rural_connection_rate_df.reset_index(drop=True, inplace=True)

            urban_treatment_rate_df = urban_treatment_rate_df_sdg = df_dmds[
                df_dmds["variable"] == "urban_treatment_rate_" + pol_scen
            ]
            urban_treatment_rate_df.reset_index(drop=True, inplace=True)

            rural_treatment_rate_df = rural_treatment_rate_df_sdg = df_dmds[
                df_dmds["variable"] == "rural_treatment_rate_" + pol_scen
            ]
            rural_treatment_rate_df.reset_index(drop=True, inplace=True)

            df_recycling = df_recycling_sdg = df_dmds[
                df_dmds["variable"] == "urban_recycling_rate_" + pol_scen
            ]
            df_recycling.reset_index(drop=True, inplace=True)

        all_rates_sdg = pd.concat(
            [
                urban_connection_rate_df_sdg,
                rural_connection_rate_df_sdg,
                urban_treatment_rate_df_sdg,
                rural_treatment_rate_df_sdg,
                df_recycling_sdg,
            ]
        )
        all_rates_sdg["variable"] = [
            x.replace("baseline", pol_scen) for x in all_rates_sdg["variable"]
        ]
        all_rates = pd.concat([all_rates_base, all_rates_sdg])
        save_path = package_data_path("water", "demands", "harmonized", context.regions)
        # save all the rates for reporting purposes
        all_rates.to_csv(save_path / "all_rates_SSP2.csv", index=False)

    # urban water demand and return. 1e-3 from mcm to km3
    urban_mw = urban_withdrawal_df.reset_index(drop=True)
    urban_mw = urban_mw.merge(
        urban_connection_rate_df.drop(columns=["variable", "time"]).rename(
            columns={"value": "rate"}
        )
    )
    urban_mw["value"] = (1e-3 * urban_mw["value"]) * urban_mw["rate"]

    dmd_df = make_df(
        "demand",
        node="B" + urban_mw["node"],
        commodity="urban_mw",
        level="final",
        year=urban_mw["year"],
        time=urban_mw["time"],
        value=urban_mw["value"],
        unit="km3/year",
    )
    urban_dis = urban_withdrawal_df.reset_index(drop=True)
    urban_dis = urban_dis.merge(
        urban_connection_rate_df.drop(columns=["variable", "time"]).rename(
            columns={"value": "rate"}
        )
    )
    urban_dis["value"] = (1e-3 * urban_dis["value"]) * (1 - urban_dis["rate"])

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + urban_dis["node"],
            commodity="urban_disconnected",
            level="final",
            year=urban_dis["year"],
            time=urban_dis["time"],
            value=urban_dis["value"],
            unit="km3/year",
        )
    )
    # rural water demand and return
    rural_mw = rual_withdrawal_df.reset_index(drop=True)
    rural_mw = rural_mw.merge(
        rural_connection_rate_df.drop(columns=["variable", "time"]).rename(
            columns={"value": "rate"}
        )
    )
    rural_mw["value"] = (1e-3 * rural_mw["value"]) * rural_mw["rate"]

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + rural_mw["node"],
            commodity="rural_mw",
            level="final",
            year=rural_mw["year"],
            time=rural_mw["time"],
            value=rural_mw["value"],
            unit="km3/year",
        )
    )

    rural_dis = rual_withdrawal_df.reset_index(drop=True)
    rural_dis = rural_dis.merge(
        rural_connection_rate_df.drop(columns=["variable", "time"]).rename(
            columns={"value": "rate"}
        )
    )
    rural_dis["value"] = (1e-3 * rural_dis["value"]) * (1 - rural_dis["rate"])

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + rural_dis["node"],
            commodity="rural_disconnected",
            level="final",
            year=rural_dis["year"],
            time=rural_dis["time"],
            value=rural_dis["value"],
            unit="km3/year",
        )
    )

    # manufactury/ industry water demand and return
    manuf_mw = industrial_withdrawals_df.reset_index(drop=True)
    manuf_mw["value"] = 1e-3 * manuf_mw["value"]

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + manuf_mw["node"],
            commodity="industry_mw",
            level="final",
            year=manuf_mw["year"],
            time=manuf_mw["time"],
            value=manuf_mw["value"],
            unit="km3/year",
        )
    )

    manuf_uncollected_wst = industrial_return_df.reset_index(drop=True)
    manuf_uncollected_wst["value"] = 1e-3 * manuf_uncollected_wst["value"]

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + manuf_uncollected_wst["node"],
            commodity="industry_uncollected_wst",
            level="final",
            year=manuf_uncollected_wst["year"],
            time=manuf_uncollected_wst["time"],
            value=-manuf_uncollected_wst["value"],
            unit="km3/year",
        )
    )

    urban_collected_wst = urban_return_df.reset_index(drop=True)
    urban_collected_wst = urban_collected_wst.merge(
        urban_treatment_rate_df.drop(columns=["variable", "time"]).rename(
            columns={"value": "rate"}
        )
    )
    urban_collected_wst["value"] = (
        1e-3 * urban_collected_wst["value"]
    ) * urban_collected_wst["rate"]

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + urban_collected_wst["node"],
            commodity="urban_collected_wst",
            level="final",
            year=urban_collected_wst["year"],
            time=urban_collected_wst["time"],
            value=-urban_collected_wst["value"],
            unit="km3/year",
        )
    )

    rural_collected_wst = rural_return_df.reset_index(drop=True)
    rural_collected_wst = rural_collected_wst.merge(
        rural_treatment_rate_df.drop(columns=["variable", "time"]).rename(
            columns={"value": "rate"}
        )
    )
    rural_collected_wst["value"] = (
        1e-3 * rural_collected_wst["value"]
    ) * rural_collected_wst["rate"]

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + rural_collected_wst["node"],
            commodity="rural_collected_wst",
            level="final",
            year=rural_collected_wst["year"],
            time=rural_collected_wst["time"],
            value=-rural_collected_wst["value"],
            unit="km3/year",
        )
    )
    urban_uncollected_wst = urban_return_df.reset_index(drop=True)
    urban_uncollected_wst = urban_uncollected_wst.merge(
        urban_treatment_rate_df.drop(columns=["variable", "time"]).rename(
            columns={"value": "rate"}
        )
    )
    urban_uncollected_wst["value"] = (1e-3 * urban_uncollected_wst["value"]) * (
        1 - urban_uncollected_wst["rate"]
    )

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + urban_uncollected_wst["node"],
            commodity="urban_uncollected_wst",
            level="final",
            year=urban_uncollected_wst["year"],
            time=urban_uncollected_wst["time"],
            value=-urban_uncollected_wst["value"],
            unit="km3/year",
        )
    )

    rural_uncollected_wst = rural_return_df.reset_index(drop=True)
    rural_uncollected_wst = rural_uncollected_wst.merge(
        rural_treatment_rate_df.drop(columns=["variable", "time"]).rename(
            columns={"value": "rate"}
        )
    )
    rural_uncollected_wst["value"] = (1e-3 * rural_uncollected_wst["value"]) * (
        1 - rural_uncollected_wst["rate"]
    )

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + rural_uncollected_wst["node"],
            commodity="rural_uncollected_wst",
            level="final",
            year=rural_uncollected_wst["year"],
            time=rural_uncollected_wst["time"],
            value=-rural_uncollected_wst["value"],
            unit="km3/year",
        )
    )
    # Add 2010 & 2015 values as historical activities to corresponding technologies
    h_act = dmd_df[dmd_df["year"].isin([2010, 2015])]

    dmd_df = dmd_df[dmd_df["year"].isin(info.Y)]
    results["demand"] = dmd_df

    # create a list of our conditions
    conditions = [
        (h_act["commodity"] == "urban_mw"),
        (h_act["commodity"] == "industry_mw"),
        (h_act["commodity"] == "rural_mw"),
        (h_act["commodity"] == "urban_disconnected"),
        (h_act["commodity"] == "rural_disconnected"),
        (h_act["commodity"] == "urban_collected_wst"),
        (h_act["commodity"] == "rural_collected_wst"),
        (h_act["commodity"] == "urban_uncollected_wst"),
        (h_act["commodity"] == "industry_uncollected_wst"),
        (h_act["commodity"] == "rural_uncollected_wst"),
    ]

    # create a list of the values we want to assign for each condition
    values = [
        "urban_t_d",
        "industry_unconnected",
        "rural_t_d",
        "urban_unconnected",
        "rural_unconnected",
        "urban_sewerage",
        "rural_sewerage",
        "urban_untreated",
        "industry_untreated",
        "rural_untreated",
    ]
    # create a new column and use np.select to assign
    # values to it using our lists as arguments
    h_act["commodity"] = np.select(conditions, values)
    h_act["value"] = h_act["value"].abs()

    hist_act = make_df(
        "historical_activity",
        node_loc=h_act["node"],
        technology=h_act["commodity"],
        year_act=h_act["year"],
        mode="M1",
        time=h_act["time"],
        value=h_act["value"],
        unit="km3/year",
    )
    results["historical_activity"] = hist_act

    h_cap = h_act[h_act["year"] >= 2015]
    h_cap = (
        h_cap.groupby(["node", "commodity", "level", "year", "unit"])["value"]
        .sum()
        .reset_index()
    )

    hist_cap = make_df(
        "historical_new_capacity",
        node_loc=h_cap["node"],
        technology=h_cap["commodity"],
        year_vtg=h_cap["year"],
        value=h_cap["value"] / 5,
        unit="km3/year",
    )

    results["historical_new_capacity"] = hist_cap

    # share constraint lower bound on urban_Water recycling
    df_share_wat = make_df(
        "share_commodity_lo",
        shares="share_wat_recycle",
        node_share="B" + df_recycling["node"],
        year_act=df_recycling["year"],
        value=df_recycling["value"],
        unit="-",
    ).pipe(
        broadcast,
        time=sub_time,
    )

    df_share_wat = df_share_wat[df_share_wat["year_act"].isin(info.Y)]
    results["share_commodity_lo"] = df_share_wat

    # rel = make_df(
    #     "relation_activity",
    #     relation="recycle_rel",
    #     node_rel="B" + df_recycling["node"],
    #     year_rel=df_recycling["year"],
    #     node_loc="B" + df_recycling["node"],
    #     technology="urban_recycle",
    #     year_act=df_recycling["year"],
    #     mode="M1",
    #     value=-df_recycling["value"],
    #     unit="-",
    # )

    # rel = rel.append(
    #     make_df(
    #         "relation_activity",
    #         relation="recycle_rel",
    #         node_rel="B" + df_recycling["node"],
    #         year_rel=df_recycling["year"],
    #         node_loc="B" + df_recycling["node"],
    #         technology="urban_sewerage",
    #         year_act=df_recycling["year"],
    #         mode="M1",
    #         value=1,
    #         unit="-",
    #     )
    # )

    # results["relation_activity"] = rel

    # rel_lo = make_df(
    #     "relation_lower",
    #     relation="recycle_rel",
    #     node_rel="B" + df_recycling["node"],
    #     value=0,
    #     unit="-",
    # ).pipe(broadcast, year_rel=info.Y)

    # results["relation_lower"] = rel_lo

    # rel_up = make_df(
    #     "relation_upper",
    #     relation="recycle_rel",
    #     node_rel="B" + df_recycling["node"],
    #     value=0,
    #     unit="-",
    # ).pipe(broadcast, year_rel=info.Y)

    # results["relation_upper"] = rel_up

    return results


def read_water_availability(context):
    """
    Reads water availability data and bias correct
    it for the historical years and no climate
    scenario assumptions.

    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'. Values
        are data frames ready for :meth:`~.Scenario.add_par`.
    """

    # Reference to the water configuration
    info = context["water build info"]
    # reading sample for assiging basins
    PATH = package_data_path(
        "water", "delineation", f"basins_by_region_simpl_{context.regions}.csv"
    )
    df_x = pd.read_csv(PATH)

    if "year" in context.time:
        # path for reading basin delineation file
        PATH = package_data_path(
            "water", "delineation", f"basins_by_region_simpl_{context.regions}.csv"
        )
        df_x = pd.read_csv(PATH)
        # Adding freshwater supply constraints
        # Reading data, the data is spatially and temprally aggregated from GHMs
        path1 = package_data_path(
            "water",
            "availability",
            f"qtot_5y_{context.RCP}_{context.REL}_{context.regions}.csv",
        )
        # Read rcp 2.6 data
        df_sw = pd.read_csv(path1)
        df_sw.drop(["Unnamed: 0"], axis=1, inplace=True)

        df_sw.index = df_x["BCU_name"]
        df_sw = df_sw.stack().reset_index()
        df_sw.columns = ["Region", "years", "value"]
        df_sw.fillna(0, inplace=True)
        df_sw.reset_index(drop=True, inplace=True)
        df_sw["year"] = pd.DatetimeIndex(df_sw["years"]).year
        df_sw["time"] = "year"
        df_sw2210 = df_sw[df_sw["year"] == 2100]
        df_sw2210["year"] = 2110
        df_sw = pd.concat([df_sw, df_sw2210])
        df_sw = df_sw[df_sw["year"].isin(info.Y)]

        # Adding groundwater supply constraints
        # Reading data, the data is spatially and temprally aggregated from GHMs
        path1 = package_data_path(
            "water",
            "availability",
            f"qr_5y_{context.RCP}_{context.REL}_{context.regions}.csv",
        )

        # Read groundwater data
        df_gw = pd.read_csv(path1)
        df_gw.drop(["Unnamed: 0"], axis=1, inplace=True)
        df_gw.index = df_x["BCU_name"]
        df_gw = df_gw.stack().reset_index()
        df_gw.columns = ["Region", "years", "value"]
        df_gw.fillna(0, inplace=True)
        df_gw.reset_index(drop=True, inplace=True)
        df_gw["year"] = pd.DatetimeIndex(df_gw["years"]).year
        df_gw["time"] = "year"
        df_gw2210 = df_gw[df_gw["year"] == 2100]
        df_gw2210["year"] = 2110
        df_gw = pd.concat([df_gw, df_gw2210])
        df_gw = df_gw[df_gw["year"].isin(info.Y)]

    else:
        # Adding freshwater supply constraints
        # Reading data, the data is spatially and temprally aggregated from GHMs
        path1 = package_data_path(
            "water",
            "availability",
            f"qtot_5y_m_{context.RCP}_{context.REL}_{context.regions}.csv",
        )
        df_sw = pd.read_csv(path1)
        df_sw.drop(["Unnamed: 0"], axis=1, inplace=True)

        df_sw.index = df_x["BCU_name"]
        df_sw = df_sw.stack().reset_index()
        df_sw.columns = ["Region", "years", "value"]
        df_sw.sort_values(["Region", "years", "value"], inplace=True)
        df_sw.fillna(0, inplace=True)
        df_sw.reset_index(drop=True, inplace=True)
        df_sw["year"] = pd.DatetimeIndex(df_sw["years"]).year
        df_sw["time"] = pd.DatetimeIndex(df_sw["years"]).month
        df_sw2210 = df_sw[df_sw["year"] == 2100]
        df_sw2210["year"] = 2110
        df_sw = pd.concat([df_sw, df_sw2210])
        df_sw = df_sw[df_sw["year"].isin(info.Y)]

        # Reading data, the data is spatially and temporally aggregated from GHMs
        path1 = package_data_path(
            "water",
            "availability",
            f"qr_5y_m_{context.RCP}_{context.REL}_{context.regions}.csv",
        )
        df_gw = pd.read_csv(path1)
        df_gw.drop(["Unnamed: 0"], axis=1, inplace=True)

        df_gw.index = df_x["BCU_name"]
        df_gw = df_gw.stack().reset_index()
        df_gw.columns = ["Region", "years", "value"]
        df_gw.sort_values(["Region", "years", "value"], inplace=True)
        df_gw.fillna(0, inplace=True)
        df_gw.reset_index(drop=True, inplace=True)
        df_gw["year"] = pd.DatetimeIndex(df_gw["years"]).year
        df_gw["time"] = pd.DatetimeIndex(df_gw["years"]).month
        df_gw2210 = df_gw[df_gw["year"] == 2100]
        df_gw2210["year"] = 2110
        df_gw = pd.concat([df_gw, df_sw2210])
        df_gw = df_gw[df_gw["year"].isin(info.Y)]

    return df_sw, df_gw


def add_water_availability(context):
    """
    Adds water supply constraints
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'. Values
        are data frames ready for :meth:`~.Scenario.add_par`.
    """

    # define an empty dictionary
    results = {}
    # Adding freshwater supply constraints
    # Reading data, the data is spatially and temprally aggregated from GHMs

    df_sw, df_gw = read_water_availability(context)

    dmd_df = make_df(
        "demand",
        node="B" + df_sw["Region"].astype(str),
        commodity="surfacewater_basin",
        level="water_avail_basin",
        year=df_sw["year"],
        time=df_sw["time"],
        value=-df_sw["value"],
        unit="km3/year",
    )

    dmd_df = dmd_df.append(
        make_df(
            "demand",
            node="B" + df_gw["Region"].astype(str),
            commodity="groundwater_basin",
            level="water_avail_basin",
            year=df_gw["year"],
            time=df_gw["time"],
            value=-df_gw["value"],
            unit="km3/year",
        )
    )

    dmd_df["value"] = dmd_df["value"].apply(lambda x: x if x <= 0 else 0)

    results["demand"] = dmd_df

    # share constraint lower bound on groundwater
    df_share = make_df(
        "share_commodity_lo",
        shares="share_low_lim_GWat",
        node_share="B" + df_gw["Region"],
        year_act=df_gw["year"],
        time=df_gw["time"],
        value=df_gw["value"]
        / (df_sw["value"] + df_gw["value"])
        * 0.95,  # 0.95 buffer factor to avoid numerical error
        unit="-",
    )

    df_share["value"] = df_share["value"].fillna(0)

    results["share_commodity_lo"] = df_share

    return results


def add_irrigation_demand(context):
    """
    Adds endogenous irrigation water demands from GLOBIOM emulator
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'. Values
        are data frames ready for :meth:`~.Scenario.add_par`.
    """
    # define an empty dictionary
    results = {}

    scen = context.get_scenario()
    # add water for irrigation from globiom
    land_out_1 = scen.par(
        "land_output", {"commodity": "Water|Withdrawal|Irrigation|Cereals"}
    )
    land_out_1["level"] = "irr_cereal"
    land_out_2 = scen.par(
        "land_output", {"commodity": "Water|Withdrawal|Irrigation|Oilcrops"}
    )
    land_out_2["level"] = "irr_oilcrops"
    land_out_3 = scen.par(
        "land_output", {"commodity": "Water|Withdrawal|Irrigation|Sugarcrops"}
    )
    land_out_3["level"] = "irr_sugarcrops"

    land_out = pd.concat([land_out_1, land_out_2, land_out_3])
    land_out["commodity"] = "freshwater"

    land_out["value"] = 1e-3 * land_out["value"]

    # take land_out edited and add as a demand in  land_input
    results["land_input"] = land_out

    return results



import logging
from pathlib import Path

import click

from message_ix_models.util.click import common_params

log = logging.getLogger(__name__)


def _modules_arg(context, param, value):
    """--module/-m: load extra reporting config from modules."""
    from . import register

    for m in filter(len, value.split(",")):
        name = register(m)
        log.info(f"Registered reporting from {name}")


@click.command(name="report")
@common_params("dry_run urls_from_file")
@click.option(
    "--config",
    "config_file",
    default="global",
    show_default=True,
    help="Path or stem for reporting config file.",
)
@click.option("--legacy", "-L", is_flag=True, help="Invoke 'legacy' reporting.")
@click.option(
    "--module",
    "-m",
    "modules",
    metavar="MODULES",
    default="",
    help="Add extra reporting for MODULES.",
    callback=_modules_arg,
)
@click.option(
    "--output",
    "-o",
    "cli_output",
    metavar="PATH",
    type=click.Path(writable=True, resolve_path=True, path_type=Path),
    help="Write output to PATH instead of console or default locations.",
)
@click.argument("key", default="message::default")
@click.pass_obj
def cli(context, config_file, legacy, cli_output, key, **kwargs):
    """Postprocess results.

    KEY defaults to the comprehensive report 'message::default', but may also be the
    name of a specific model quantity, e.g. 'output'.

    --config can give either the absolute path to a reporting configuration file, or
    the stem (i.e. name without .yaml extension) of a file in data/report.

    With --urls-from-file, read multiple Scenario identifiers from FILE, and report each
    one. In this usage, --output-path may only be a directory.
    """
    from copy import deepcopy

    from message_ix_models.util._logging import mark_time

    from . import report
    from .config import Config

    # Update the reporting configuration from command-line parameters
    context.report = Config(
        from_file=config_file, key=key, cli_output=cli_output, _legacy=legacy
    )

    # Prepare a list of Context objects, each referring to one Scenario.
    contexts = []

    # - If --urls-from-file was given, then `context.scenarios` will contain a list of
    #   ScenarioInfo objects that point to the platform and (model, scenario, version)
    #   identifiers.
    # - Otherwise, the user gave identifiers for a single Scenario to the top-level CLI
    #   (--url/--platform/--model/--scenario/--version) and these are stored in
    #   `context.platform_info` and `context.scenario_info`.
    for si in context.scenarios or [context.scenario_info]:
        ctx = deepcopy(context)
        ctx.platform_info = dict(
            name=getattr(
                si, "platform_name", context.platform_info.get("name", "default")
            )
        )
        ctx.scenario_info = dict(si)
        contexts.append(ctx)

    for ctx in contexts:
        mark_time()
        report(ctx)

    mark_time()



import pytest
from genno import Computer

from message_ix_models.project.advance.data import ADVANCE  # noqa: F401
from message_ix_models.tools.exo_data import prepare_computer


class TestADVANCE:
    @pytest.mark.parametrize(
        "source_kw, dimensionality",
        (
            (
                dict(
                    measure="Transport|Service demand|Road|Freight",
                    model="MESSAGE",
                    scenario="ADV3TRAr2_Base",
                ),
                {"[length]": 1, "[mass]": 1, "[time]": -1},
            ),
            (
                dict(
                    measure="Transport|Service demand|Road|Passenger|LDV",
                    model="MESSAGE",
                    scenario="ADV3TRAr2_Base",
                ),
                {"[length]": 1, "[passenger]": 1, "[time]": -1},
            ),
            # Excess keyword arguments
            pytest.param(
                dict(measure="GDP", model="not a model", foo="bar"),
                None,
                marks=pytest.mark.xfail(raises=ValueError),
            ),
        ),
    )
    @pytest.mark.parametrize("regions, N_n", (("ADVANCE", 7), ("R12", 4)))
    def test_prepare_computer(
        self, test_context, source_kw, dimensionality, regions, N_n
    ):
        source = "ADVANCE"
        test_context.model.regions = regions

        c = Computer()

        keys = prepare_computer(test_context, c, source, source_kw)

        # Preparation of data runs successfully
        result = c.get(keys[0])

        # Data have the expected dimensions
        assert ("n", "y") == result.dims

        # Data units have the expected dimensionality
        assert dimensionality == result.units.dimensionality

        # Data are complete
        assert N_n == len(result.coords["n"])
        assert 14 == len(result.coords["y"])



import logging
import re

import pytest
from sdmx.model.v21 import Annotation, Code

from message_ix_models.util.sdmx import eval_anno, make_enum, read


def test_eval_anno(caplog):
    c = Code()

    assert None is eval_anno(c, "foo")

    c.annotations.append(Annotation(id="foo", text="bar baz"))

    with caplog.at_level(logging.DEBUG, logger="message_ix_models"):
        assert "bar baz" == eval_anno(c, "foo")

    assert re.fullmatch(
        r"Could not eval\('bar baz'\): .* \(<string>, line 1\)", caplog.messages[0]
    )

    c.annotations.append(Annotation(id="qux", text="3 + 4"))

    assert 7 == eval_anno(c, id="qux")


def test_make_enum():
    """:func:`.make_enum` works with :class:`~enum.Flag` and subclasses."""
    from enum import Flag, IntFlag

    E = make_enum("ICONICS:SSP(2017)", base=Flag)

    # Values are bitwise flags
    assert not isinstance(E["1"], int)

    # Expected length
    assert 2 ** (len(E) - 1) == list(E)[-1].value

    # Flags can be combined
    flags = E["1"] | E["2"]
    assert E["1"] & flags
    assert E["2"] & flags
    assert not (E["3"] & flags)

    # Similar, with IntFlag
    E = make_enum("IIASA_ECE:AGENCIES(0.1)", base=IntFlag)

    # Values are ints
    assert isinstance(E["IIASA_ECE"], int)

    # Expected length
    assert 2 ** (len(E) - 1) == list(E)[-1].value


@pytest.mark.parametrize(
    "urn, expected",
    (
        ("ICONICS:SSP(2017)", "Codelist=ICONICS:SSP(2017)"),
        ("ICONICS:SSP(2024)", "Codelist=ICONICS:SSP(2024)"),
        ("SSP(2017)", "Codelist=ICONICS:SSP(2017)"),
        ("SSP(2024)", "Codelist=ICONICS:SSP(2024)"),
        ("SSP", "Codelist=ICONICS:SSP(2017)"),
        ("AGENCIES", "AgencyScheme=IIASA_ECE:AGENCIES(0.1)"),
    ),
)
def test_read0(urn, expected):
    obj = read(urn)
    assert expected in obj.urn


def test_read1():
    SSPS = read("ssp")

    # Identify an SSP by matching strings in its name
    code0 = next(filter(lambda c: "2" in repr(c), iter(SSPS)))
    code1 = next(filter(lambda c: "SSP2" in repr(c), iter(SSPS)))
    code2 = next(filter(lambda c: "middle of the road" in repr(c).lower(), iter(SSPS)))

    assert code0 is code1 is code2

    with pytest.raises(FileNotFoundError):
        read("foo")



"""Prepare data for water use for cooling & energy technologies."""

import pandas as pd
from message_ix import make_df

from message_ix_models.util import broadcast, package_data_path


# water & electricity for irrigation
def add_irr_structure(context):
    """Add irrigation withdrawal infrastructure
    The irrigation demands are added in
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'.
        Values are data frames ready for :meth:`~.Scenario.add_par`.
        Years in the data include the model horizon indicated by
        ``context["water build info"]``, plus the additional year 2010.
    """

    # define an empty dictionary
    results = {}

    # reading basin_delineation
    FILE2 = f"basins_by_region_simpl_{context.regions}.csv"
    PATH = package_data_path("water", "delineation", FILE2)
    df_node = pd.read_csv(PATH)
    # Assigning proper nomenclature
    df_node["node"] = "B" + df_node["BCU_name"].astype(str)
    df_node["mode"] = "M" + df_node["BCU_name"].astype(str)
    if context.type_reg == "country":
        df_node["region"] = context.map_ISO_c[context.regions]
    else:
        df_node["region"] = f"{context.regions}_" + df_node["REGION"].astype(str)

    # Reference to the water configuration
    info = context["water build info"]

    # probably can be removed
    year_wat = [2010, 2015]
    year_wat.extend(info.Y)

    inp = make_df(
        "input",
        technology="irrigation_cereal",
        value=1,
        unit="-",
        level="water_supply",
        commodity="freshwater",
        mode="M1",
        time="year",
        time_origin="year",
        node_origin=df_node["region"],
        node_loc=df_node["region"],
    ).pipe(broadcast, year_vtg=info.Y)

    inp = inp.append(
        make_df(
            "input",
            technology="irrigation_oilcrops",
            value=1,
            unit="-",
            level="water_supply",
            commodity="freshwater",
            mode="M1",
            time="year",
            time_origin="year",
            node_origin=df_node["region"],
            node_loc=df_node["region"],
        ).pipe(broadcast, year_vtg=info.Y)
    )

    inp = inp.append(
        make_df(
            "input",
            technology="irrigation_sugarcrops",
            value=1,
            unit="-",
            level="water_supply",
            commodity="freshwater",
            mode="M1",
            time="year",
            time_origin="year",
            node_origin=df_node["region"],
            node_loc=df_node["region"],
        ).pipe(broadcast, year_vtg=info.Y)
    )
    # year_act = year_vts for tecs with 1 time-step lifetime
    inp["year_act"] = inp["year_vtg"]

    # Electricity values per unit of irrigation water supply
    # Reference: Evaluation of Water and Energy Use in
    # Pressurized Irrigation Networks in Southern Spain
    # Diaz et al. 2011 https://ascelibrary.org/
    # doi/10.1061/%28ASCE%29IR.1943-4774.0000338
    # Low Value :0.04690743
    # Average Value :0.101598174
    # High Value : 0.017123288

    # inp = inp.append(
    #     make_df(
    #         "input",
    #         technology="irrigation_sugarcrops",
    #         value=0.04690743,
    #         unit="-",
    #         level="final",
    #         commodity="electr",
    #         mode="M1",
    #         time="year",
    #         time_origin="year",
    #         node_origin=df_node["region"],
    #         node_loc=df_node["region"],
    #     ).pipe(broadcast, year_vtg=year_wat, year_act=year_wat)
    # )
    #
    # inp = inp.append(
    #     make_df(
    #         "input",
    #         technology="irrigation_oilcrops",
    #         value=0.04690743,
    #         unit="-",
    #         level="final",
    #         commodity="electr",
    #         mode="M1",
    #         time="year",
    #         time_origin="year",
    #         node_origin=df_node["region"],
    #         node_loc=df_node["region"],
    #     ).pipe(broadcast, year_vtg=year_wat, year_act=year_wat)
    # )
    #
    # inp = inp.append(
    #     make_df(
    #         "input",
    #         technology="irrigation_cereal",
    #         value=0.04690743,
    #         unit="-",
    #         level="final",
    #         commodity="electr",
    #         mode="M1",
    #         time="year",
    #         time_origin="year",
    #         node_origin=df_node["region"],
    #         node_loc=df_node["region"],
    #     ).pipe(broadcast, year_vtg=year_wat, year_act=year_wat)
    # )
    # inp.loc[(inp['node_loc'] == 'R11_SAS') &
    #         (inp['commodity'] == 'electr'),
    #         "value",
    # ] *= 0.00004690743

    results["input"] = inp

    irr_out = make_df(
        "output",
        technology="irrigation_cereal",
        value=1,
        unit="km3/year",
        level="irr_cereal",
        commodity="freshwater",
        mode="M1",
        time="year",
        time_dest="year",
        node_loc=df_node["region"],
        node_dest=df_node["region"],
    ).pipe(broadcast, year_vtg=info.Y)

    irr_out = irr_out.append(
        make_df(
            "output",
            technology="irrigation_sugarcrops",
            value=1,
            unit="km3/year",
            level="irr_sugarcrops",
            commodity="freshwater",
            mode="M1",
            time="year",
            time_dest="year",
            node_loc=df_node["region"],
            node_dest=df_node["region"],
        ).pipe(
            broadcast,
            year_vtg=info.Y,
        )
    )

    irr_out = irr_out.append(
        make_df(
            "output",
            technology="irrigation_oilcrops",
            value=1,
            unit="km3/year",
            level="irr_oilcrops",
            commodity="freshwater",
            mode="M1",
            time="year",
            time_dest="year",
            node_loc=df_node["region"],
            node_dest=df_node["region"],
        ).pipe(
            broadcast,
            year_vtg=info.Y,
        )
    )

    irr_out["year_act"] = irr_out["year_vtg"]

    results["output"] = irr_out

    return results



from dataclasses import dataclass, field
from typing import Type

import pytest

from message_ix_models.util.config import ConfigHelper


class TestConfigHelper:
    @pytest.fixture
    def cls(self) -> Type:
        """A class which inherits from ConfigHelper."""

        @dataclass
        class Config(ConfigHelper):
            foo_1: int = 1
            foo_2: str = ""
            foo_3: bool = True

        return Config

    @pytest.fixture
    def cls2(self, cls) -> Type:
        """A class with an attribute."""

        @dataclass
        class Config3:
            """NOT a ConfigHelper subclass."""

            baz_1: int = 1
            baz_2: int = 2

        @dataclass
        class Config2(ConfigHelper):
            bar_1: float = 0.01
            subconfig_a: cls = field(default_factory=cls)  # type: ignore [valid-type]
            subconfig_b: Config3 = field(default_factory=Config3)

        return Config2

    @pytest.fixture
    def c(self, cls):
        return cls(foo_1=99, foo_2="bar", foo_3=False)

    @pytest.fixture
    def c2(self, cls, cls2):
        result = cls2(bar_1=3.14, subconfig_a=cls(foo_1=99, foo_2="bar", foo_3=False))
        result.subconfig_b.baz_1 = 2
        result.subconfig_b.baz_2 = 1
        return result

    def test_canonical_name(self, cls):
        assert "foo_1" == cls._canonical_name("foo 1")
        assert "foo_2" == cls._canonical_name("foo-2")
        assert "foo_3" == cls._canonical_name("foo_3")
        assert None is cls._canonical_name("foo 4")

    def test_from_dict(self, cls, c):
        values = {"foo 1": 99, "foo-2": "bar", "foo_3": False}
        assert c == cls.from_dict({"foo 1": 99, "foo-2": "bar", "foo_3": False})

        values.update(foo_4=3.14)
        with pytest.raises(ValueError):
            cls.from_dict(values)

    def test_read_file(self, caplog, tmp_path, cls, cls2, c, c2):
        # Write a YAML snippet to file
        yaml_path = tmp_path.joinpath("config.yaml")
        yaml_path.write_text(
            """
foo 1: 99

foo-2: bar

foo_3: false

foo_4: 3.14
            """
        )

        obj1 = cls()
        # Method runs
        obj1.read_file(yaml_path, fail=False)
        # Values are read
        assert c == obj1, obj1
        # Messages are logged
        assert [
            "Config has no attribute for file section 'foo_4'; ignored"
        ] == caplog.messages
        caplog.clear()

        yaml_path.write_text(
            """
bar_1: 3.14
subconfig a:
  foo 1: 99
  foo-2: bar
  foo_3: false
subconfig-b:  # No name manipulation for subkeys here
  baz_1: 2
  baz_2: 1
            """
        )

        obj2 = cls2()
        # Method runs
        obj2.read_file(yaml_path, fail=False)
        # Values are read
        assert c2 == obj2, obj2

        json_path = tmp_path.joinpath("config.json")
        json_path.write_text(
            """{
"foo 1": 99,
"foo-2": "bar",
"foo_3": false
}           """
        )

        obj3 = cls()
        # Method runs
        obj3.read_file(json_path)
        # Values are read
        assert c == obj3, obj3

        obj4 = cls()
        with pytest.raises(NotImplementedError):
            obj4.read_file(yaml_path.with_suffix(".xlsx"))

    def test_replace(self, c):
        result = c.replace(foo_2="baz")
        assert result is not c
        assert "baz" == result.foo_2

    def test_update(self, c):
        """:meth:`.ConfigHelper.update` raises AttributeError."""
        with pytest.raises(AttributeError):
            c.update(foo_4="")



"""Cache data for expensive operations.

This module extends :class:`genno.caching.Encoder` to handle classes common in
:mod:`message_ix_models`, so these can be used as arguments to cached functions and
included in the computed cache key:

- :class:`sdmx.model.IdentifiableArtefact`, including :class:`.Code`: hashed as
  their string representation / ID.
- :class:`ixmp.Platform`, :class:`xarray.Dataset`: ignored, with a warning logged.
- :class:`ScenarioInfo`: only the :attr:`~ScenarioInfo.set` entries are hashed.

"""
import json
import logging
from dataclasses import asdict, is_dataclass
from typing import Callable

import genno.caching
import ixmp
import sdmx.model
import xarray as xr

from .context import Context
from .scenarioinfo import ScenarioInfo

log = logging.getLogger(__name__)


#: Controls whether cached data is returned for functions decorated with
#: :func:`.cached`. Set to :obj:`True` to force reload.
SKIP_CACHE = False

# Paths already logged, to decrease verbosity
PATHS_SEEN = set()


# Show genno how to hash function arguments seen in message_ix_models


@genno.caching.Encoder.register
def _sdmx_identifiable(o: sdmx.model.IdentifiableArtefact):
    return str(o)


@genno.caching.Encoder.register
def _dataclass(o: object):
    return (
        asdict(o)
        if (is_dataclass(o) and not isinstance(o, type))
        else json.JSONEncoder().default(o)
    )


@genno.caching.Encoder.register
def _si(o: ScenarioInfo):
    return dict(o.set)


genno.caching.Encoder.ignore(xr.Dataset, ixmp.Platform)


def cached(func: Callable) -> Callable:
    """Decorator to cache the return value of a function `func`.

    On a first call, the data requested is returned and also cached under
    :meth:`.Context.get_cache_path`. On subsequent calls, if the cache exists, it is
    used instead of calling the (possibly slow) `func`.

    When :data:`.SKIP_CACHE` is true, `func` is always called.

    See also
    --------
    :doc:`genno:cache` in the :mod:`genno` documentation
    """
    # Determine and create the cache path
    cache_path = Context.get_instance(-1).get_cache_path()
    cache_path.mkdir(exist_ok=True, parents=True)

    if cache_path not in PATHS_SEEN:
        log.debug(f"{func.__name__}() will cache in {cache_path}")
        PATHS_SEEN.add(cache_path)

    # Use the genno internals to wrap the function.
    cached_load = genno.caching.decorate(
        func, cache_path=cache_path, cache_skip=SKIP_CACHE
    )

    if cached_load.__doc__ is not None:
        # Determine the indent
        line = cached_load.__doc__.split("\n")[-1]
        indent = len(line) - len(line.lstrip(" "))

        # Add a note that the results are cached
        cached_load.__doc__ += (
            f"\n\n{' ' * indent}Data returned by this function is cached using "
            ":func:`.cached`; see also :data:`.SKIP_CACHE`."
        )

    return cached_load



"""
This script aggregates the global gridded data to any scale. The following
script specifically aggregates global gridded hydrological data onto the basin
 mapping used in the nexus module.
"""

import os

import numpy as np
import pandas as pd

# variable, for detailed symbols, refer to ISIMIP2b documentation
variables = ["qtot", "dis", "qr"]  # total runoff  # discharge  # groundwater run

var = "qr"


GCMs = ["mmean", "dry"]
GCM = "mmean"  # 'dry'choose driest GCM else 'mmean' for multi model mean
# else mmmean
iso3 = "ZMB"
isimip = "3b"

# climate model
if isimip == "2b":
    climmodels = ["gfdl-esm2m", "hadgem2-es", "ipsl-cm5a-lr", "miroc5"]
    climmodel = "gfdl-esm2m"
    # climate forcing
    scenarios = ["rcp26", "rcp60"]
    scen = "rcp26"
    # This is an internal IIASA directory path
    wd1 = os.path.join("p:", "ene.model", "NEST", "hydrological_data_agg") + os.sep
    wd = os.path.join("p:", "watxene", "ISIMIP", "ISIMIP2b", "output", "LPJmL") + os.sep
    wd2 = os.path.join("p:", "ene.model", "NEST", "hydrology", "processed_nc4") + os.sep
else:
    climmodels = [
        "gfdl-esm4",
        "ipsl-cm6a-lr",
        "mpi-esm1-2-hr",
        "mri-esm2-0",
        "ukesm1-0-ll",
    ]
    cl = "gfdl-esm4"
    scenarios = ["ssp126", "ssp370", "ssp585"]
    scen = "ssp126"
    wd1 = os.path.join("p:", "ene.model", "NEST", "hydrological_data_agg") + os.sep
    wd11 = (
        os.path.join("p:", "ene.model", "NEST", iso3, "hydrology", "isimip3B_CWatM")
        + os.sep
    )


qtot_7p0_gfdl = pd.read_csv(wd11 + f"{var}_monthly_gfdl-esm4_ssp370_future.csv").drop(
    [
        "Unnamed: 0",
        "NAME",
        "BASIN",
        "BASIN_ID",
        "area",
        "area_km2",
        "X",
        "REGION",
        "BCU_name",
    ],
    axis=1,
)

new_cols = pd.date_range("2015-01-01", periods=1032, freq="M")
qtot_7p0_gfdl.columns = new_cols


qtot_2p6_gfdl = pd.read_csv(wd11 + f"{var}_monthly_gfdl-esm4_ssp126_future.csv").drop(
    [
        "Unnamed: 0",
        "NAME",
        "BASIN",
        "BASIN_ID",
        "area",
        "area_km2",
        "X",
        "REGION",
        "BCU_name",
    ],
    axis=1,
)

new_cols = pd.date_range("2015-01-01", periods=1032, freq="M")
qtot_2p6_gfdl.columns = new_cols


qtot_2p6_gfdl_avg = qtot_2p6_gfdl[pd.date_range("2015-01-01", periods=192, freq="M")]
qtot_7p0_gfdl_avg = qtot_7p0_gfdl[pd.date_range("2015-01-01", periods=192, freq="M")]

val_2020 = (
    qtot_7p0_gfdl_avg.groupby(qtot_7p0_gfdl_avg.columns.month, axis=1).mean()
    + qtot_2p6_gfdl_avg.groupby(qtot_2p6_gfdl_avg.columns.month, axis=1).mean()
) / 2
val_2020_annual = val_2020.mean(axis=1)

delta60 = (
    qtot_7p0_gfdl_avg.groupby(qtot_7p0_gfdl_avg.columns.month, axis=1).mean() - val_2020
)
delta26 = (
    qtot_2p6_gfdl_avg.groupby(qtot_2p6_gfdl_avg.columns.month, axis=1).mean() - val_2020
)


def bias_correction(df):
    """ias corrects the data such that 2020 value is same
    for both scenario and then apply
    Input
    ----------
    df : raw input monthly data
    Returns
    -------
    df : bias corrected monthly data with also replacing 5 year timestep average
    df_monthly: monthly bias corrected data
    df_5y_m: bias corrected 5 y monthly
    df_5y: bias corrected 5 y average
    """
    # Starting value of delat is 1
    # it will reduce to zero with each 5 year timestep
    # till 2045 with a difference of 0.2.
    # This means the bias correction will fade away till 2045
    delta_multiply = 1

    for year in np.arange(2020, 2105, 5):
        # for 2020 delta all scenario data frame will be same
        if year == 2020:
            df[pd.date_range("2020-01-01", periods=12, freq="M")] = val_2020

        else:
            # for delta years after 2020
            if delta_multiply > 0.1:
                for i in np.arange(4, -1, -1):
                    if i == 4:
                        delta60.columns = pd.date_range(
                            str(year - i) + "-01-01", periods=12, freq="M"
                        )
                        final_temp = df[
                            pd.date_range(
                                str(year - i) + "-01-01", periods=12, freq="M"
                            )
                        ] - (delta_multiply * delta60)

                    else:
                        delta60.columns = pd.date_range(
                            str(year - i) + "-01-01", periods=12, freq="M"
                        )
                        temp = df[
                            pd.date_range(
                                str(year - i) + "-01-01", periods=12, freq="M"
                            )
                        ] - (delta_multiply * delta60)

                        final_temp = pd.concat((final_temp, temp), axis=1)

                df_monthly = final_temp
                df[
                    pd.date_range(str(year) + "-01-01", periods=12, freq="M")
                ] = final_temp.groupby(final_temp.columns.month, axis=1).mean()
                # 5 year monthly data
                df_5y_m = df[df.columns[df.columns.year.isin(years)]]
                # 5 year annual
                # 50th quantile - q50
                df_q50 = (
                    df_monthly.rolling(240, min_periods=1, axis=1)
                    .quantile(0.5, interpolation="linear")
                    .resample("5Y", axis=1)
                    .mean()
                )
                df_q50["2020-12-31"] = val_2020_annual
                # 70th quantile - q70
                df_q70 = (
                    df_monthly.rolling(240, min_periods=1, axis=1)
                    .quantile(0.3, interpolation="linear")
                    .resample("5Y", axis=1)
                    .mean()
                )
                df_q70["2020-12-31"] = val_2020_annual
                # 90th quantile - q90
                df_q90 = (
                    df_monthly.rolling(240, min_periods=1, axis=1)
                    .quantile(0.1, interpolation="linear")
                    .resample("5Y", axis=1)
                    .mean()
                )

                df_q90["2020-12-31"] = val_2020_annual

                delta_multiply -= 0.2

            else:
                # after detlta years
                temp_daterange = pd.date_range(
                    str(year - 4) + "-01-01", periods=60, freq="M"
                )
                # Monthly Bias corrected data
                df_monthly = df

                df[pd.date_range(str(year) + "-01-01", periods=12, freq="M")] = (
                    df[temp_daterange]
                    .groupby(df[temp_daterange].columns.month, axis=1)
                    .mean()
                )
                # 5 year monthly
                df_5y_m = df[df.columns[df.columns.year.isin(years)]]
                # 5 year annual
                # 50th quantile - q50
                df_q50 = (
                    df_monthly.rolling(240, min_periods=1, axis=1)
                    .quantile(0.5, interpolation="linear")
                    .resample("5Y", axis=1)
                    .mean()
                )
                df_q50["2020-12-31"] = val_2020_annual
                # 70th quantile - q70
                df_q70 = (
                    df_monthly.rolling(240, min_periods=1, axis=1)
                    .quantile(0.3, interpolation="linear")
                    .resample("5Y", axis=1)
                    .mean()
                )
                df_q70["2020-12-31"] = val_2020_annual
                # 90th quantile - q90
                df_q90 = (
                    df_monthly.rolling(240, min_periods=1, axis=1)
                    .quantile(0.1, interpolation="linear")
                    .resample("5Y", axis=1)
                    .mean()
                )

                df_q90["2020-12-31"] = val_2020_annual

    return df, df_monthly, df_5y_m, df_q50, df_q70, df_q90


years = np.arange(2015, 2105, 5)

df_26, df_monthly_26, df_5y_m_26, df_q50_26, df_q70_26, df_q90_26 = bias_correction(
    qtot_2p6_gfdl
)

df_70, df_monthly_70, df_5y_m_70, df_q50_70, df_q70_70, df_q90_70 = bias_correction(
    qtot_7p0_gfdl
)


df_monthly_26.to_csv(wd11 + f"/{var}_monthly_2p6_{iso3}.csv")
df_5y_m_26.to_csv(wd11 + f"/{var}_5y_m_2p6_low_{iso3}.csv")
df_q50_26.to_csv(wd11 + f"/{var}_5y_2p6_low_{iso3}.csv")
df_q70_26.to_csv(wd11 + f"/{var}_5y_2p6_med_{iso3}.csv")
df_q90_26.to_csv(wd11 + f"/{var}_5y_2p6_high_{iso3}.csv")

df_monthly_70.to_csv(wd11 + f"/{var}_monthly_7p0_{iso3}.csv")
df_5y_m_70.to_csv(wd11 + f"/{var}_5y_m_7p0_low_{iso3}.csv")
df_q50_70.to_csv(wd11 + f"/{var}_5y_7p0_low_{iso3}.csv")
df_q70_70.to_csv(wd11 + f"/{var}_5y_7p0_med_{iso3}.csv")
df_q90_70.to_csv(wd11 + f"/{var}_5y_7p0_high_{iso3}.csv")

# No climate scenarios
df_q50_70.apply(lambda x: val_2020_annual)
for y in years:
    df_5y_m_70[pd.date_range(f"{y}-01-01", periods=12, freq="M")] = val_2020
df_5y_m_70.to_csv(wd11 + f"/{var}_5y_m_no_climate_low_{iso3}.csv")
df_q50_70.apply(lambda x: val_2020_annual).to_csv(
    wd11 + f"/{var}_5y_no_climate_low_{iso3}.csv"
)
df_q70_70.apply(lambda x: val_2020_annual).to_csv(
    wd11 + f"/{var}_5y_no_climate_med_{iso3}.csv"
)
df_q90_70.apply(lambda x: val_2020_annual).to_csv(
    wd11 + f"/{var}_5y_no_climate_high_{iso3}.csv"
)
# Environmental Flow
df_env = df_monthly_70
col_end = None
# df = data.iloc[:,5:]
for z in range(len(df_env.columns) // 12):
    col_start = 0 if z == 0 else col_end  # start col number
    col_end = (z + 1) * 12  # ending col number
    temp = df_env.iloc[:, col_start:col_end]  # assigning relevant data

    col_names = temp.columns  # getting column names
    MAF = temp.mean(axis=1)  # take mean acorss

    for j in range(len(temp.columns)):
        temp.iloc[:, j] = np.where(
            temp.iloc[:, j] > 0.8 * MAF,
            temp.iloc[:, j] * 0.2,
            np.where(
                (temp.iloc[:, j] > 0.4 * MAF) & (temp.iloc[:, j] <= 0.8 * MAF),
                temp.iloc[:, j] * 0.45,
                temp.iloc[:, j] * 0.6,
            ),
        )

    if z == 0:
        eflow = temp
    else:
        eflow = pd.concat((eflow, temp), axis=1)

    eflow = eflow.abs()

# Convert to 5 year annual values
eflow_5y = eflow.resample("5Y", axis=1).mean()
eflow_5y.to_csv(wd11 + f"e-flow_7p0_{iso3}.csv")

for year in np.arange(2020, 2105, 5):
    for i in np.arange(4, -1, -1):
        if i == 4:
            final_temp = eflow[
                pd.date_range(str(year - i) + "-01-01", periods=12, freq="M")
            ]

        else:
            final_temp = eflow[
                pd.date_range(str(year - i) + "-01-01", periods=12, freq="M")
            ]

    df_monthly = final_temp
    eflow[
        pd.date_range(str(year) + "-01-01", periods=12, freq="M")
    ] = final_temp.groupby(final_temp.columns.month, axis=1).mean()

    eflow_5y_m = eflow[eflow.columns[eflow.columns.year.isin(years)]]
    eflow_5y_m.to_csv(wd11 + f"e-flow_5y_m_7p0_{iso3}.csv")

val_2020_eflow = eflow_5y_m[pd.date_range("2020-01-01", periods=12, freq="M")].values()
val_2020_eflowy = eflow_5y["2020-12-31"]
eflow_5y.apply(lambda x: val_2020_eflowy).to_csv(wd11 + f"e-flow_no_climate_{iso3}.csv")

for y in years:
    eflow_5y_m[pd.date_range(f"{y}-01-01", periods=12, freq="M")] = val_2020_eflow

eflow_5y_m.to_csv(wd11 + f"e-flow_5y_m_no_climate_{iso3}.csv")



"""Tools for working with IAMC-structured data."""
from typing import Optional

import pandas as pd
import sdmx.model.v21 as m
from sdmx.message import StructureMessage


def describe(data: pd.DataFrame, extra: Optional[str] = None) -> StructureMessage:
    """Generate SDMX structure information from `data` in IAMC format.

    Parameters
    ----------
    data :
        Data in "wide" or "long" IAMC format.
    extra : str, optional
        Extra text added to the description of each Codelist.

    Returns
    -------
    sdmx.message.StructureMessage
        The message contains one :class:`.Codelist` for each of the MODEL, SCENARIO,
        REGION, VARIABLE, and UNIT dimensions. Codes for the VARIABLE code list have
        annotations with :py:`id="preferred-unit-measure"` that give the corresponding
        UNIT Code(s) that appear with each VARIABLE.
    """

    sm = StructureMessage()

    def _cl(dim: str) -> m.Codelist:
        result = m.Codelist(
            id=dim,
            description=f"Codes appearing in the {dim!r} dimension of "
            + (extra or "data")
            + ".",
            is_final=True,
            is_external_reference=False,
        )
        sm.add(result)
        return result

    for dim in ("MODEL", "SCENARIO", "REGION"):
        cl = _cl(dim)
        for value in sorted(data[dim].unique()):
            cl.append(m.Code(id=value))

    # Handle "VARIABLE" and "UNIT" jointly
    dims = ["VARIABLE", "UNIT"]
    cl_variable = _cl("VARIABLE")
    cl_unit = _cl("UNIT")
    for variable, group_data in (
        data[dims].sort_values(dims).drop_duplicates().groupby("VARIABLE")
    ):
        group_units = group_data["UNIT"].unique()
        cl_variable.append(
            m.Code(
                id=variable,
                annotations=[
                    m.Annotation(
                        id="preferred-unit-measure", text=", ".join(group_units)
                    )
                ],
            )
        )
        for unit in group_units:
            try:
                cl_unit.append(m.Code(id=unit))
            except ValueError:
                pass

    return sm



import os

import click
import pytest

from message_ix_models.testing import bare_res, not_ci


def test_bare_res_no_request(test_context):
    """:func:`bare_res` works with `request` = :obj:`None`."""
    bare_res(None, test_context, solved=False)


def test_bare_res_solved(request, test_context):
    """:func:`bare_res` works with `solve` = :obj:`True`.

    This test can be removed once this feature of the test function is used by another
    test.
    """
    bare_res(request, test_context, solved=True)


def test_cli_runner(mix_models_cli):
    with pytest.raises(click.exceptions.UsageError, match="No such command 'foo'"):
        mix_models_cli.assert_exit_0(["foo", "bar"])


@not_ci(reason="foo", action="skip")
def test_not_ci_skip():
    """Test not_ci(action="skip")."""


@not_ci(reason="foo", action="xfail")
def test_not_ci_xfail():
    """Test not_ci(action="skip")."""
    assert "GITHUB_ACTIONS" not in os.environ



import logging

import click

from message_ix_models.model.structure import get_codes
from message_ix_models.util.click import common_params

log = logging.getLogger(__name__)


# allows to activate water module
@click.group("water-ix")
@common_params("regions")
@click.option("--time", help="Manually defined time")
@click.pass_obj
def cli(context, regions, time):
    """MESSAGEix-Water and Nexus variant."""
    water_ini(context, regions, time)


def water_ini(context, regions, time):
    """Add components of the MESSAGEix-Nexus module

    This function modifies model name & scenario name
    and verifies the region setup
    Parameters
    ----------
    context : `class`:message.Context
        Information about target Scenario.
    regions : str (if not defined already in context.regions)
        Specifies what region definition is used ['R11','R12','ISO3']
    """

    from .utils import read_config

    # Ensure water model configuration is loaded
    read_config(context)
    if not context.scenario_info:
        context.scenario_info.update(
            dict(model="ENGAGE_SSP2_v4.1.7", scenario="baseline_clone_test")
        )
    context.output_scenario = context.scenario_info["scenario"]
    context.output_model = context.scenario_info["model"]

    # Handle --regions; use a sensible default for MESSAGEix-Nexus
    if regions:
        log.info(f"Regions choice {regions}")
        if regions in ["R14", "R32", "RCP"]:
            log.warning(
                "the MESSAGEix-Nexus module might not be compatible"
                "with your 'regions' choice"
            )
    else:
        log.info("Use default --regions=R11")
        regions = "R11"
    # add an attribute to distinguish country models
    if regions in ["R11", "R12", "R14", "R32", "RCP"]:
        context.type_reg = "global"
    else:
        context.type_reg = "country"
    context.regions = regions

    # create a mapping ISO code :
    # region name, for other scripts
    # only needed for 1-country models
    nodes = get_codes(f"node/{context.regions}")
    nodes = list(map(str, nodes[nodes.index("World")].child))
    if context.type_reg == "country":
        map_ISO_c = {context.regions: nodes[0]}
        context.map_ISO_c = map_ISO_c
        log.info(f"mapping {context.map_ISO_c[context.regions]}")

    # deinfe the timestep
    if not time:
        sc_ref = context.get_scenario()
        time = sc_ref.set("time")
        sub_time = list(time[time != "year"])
        if len(sub_time) == 0:
            context.time = ["year"]

        else:
            context.time = sub_time
    else:
        context.time = [time]
    log.info(f"Using the following time-step for the water module: {context.time}")

    # setting the time information in context


_RCPS = ["no_climate", "6p0", "2p6", "7p0"]
_REL = ["low", "med", "high"]


@cli.command("nexus")
@click.pass_obj
@click.option("--rcps", default="6p0", type=click.Choice(_RCPS))
@click.option("--rels", default="low", type=click.Choice(_REL))
@click.option(
    "--sdgs",
    default="baseline",
    help="Defines if and what water SDG measures are activated",
)
@click.option(
    "--macro",
    is_flag=True,
    help="Defines whether the model solves with macro",
)
@common_params("regions")
def nexus_cli(context, regions, rcps, sdgs, rels, macro=False):
    """
    Add basin structure connected to the energy sector and
    water balance linking different water demands to supply.
    """

    nexus(context, regions, rcps, sdgs, rels, macro)


def nexus(context, regions, rcps, sdgs, rels, macro=False):
    """Add basin structure connected to the energy sector and
    water balance linking different water demands to supply.

    Use the --url option to specify the base scenario.

    Parameters
    ----------
    context : `class`:message.Context
        Information about target Scenario.
    regions : str (if not defined already in context.regions)
        Specifies what region definition is used ['R11','R12','ISO3']
    RCP : str
        Specifies the climate scenario used ['no_climate','6p0','2p6']
    SDG : Str
        Defines if and what water SDG measures are activated
    REL: str
        Specifies the reliability of hydrological data ['low','mid','high']
    """
    # add input information to the class context
    context.nexus_set = "nexus"
    if not context.regions:
        context.regions = regions

    context.RCP = rcps
    context.SDG = sdgs
    context.REL = rels

    log.info(f"RCP assumption is {context.RCP}. SDG is {context.SDG}")

    from .build import main as build

    # Determine the output scenario name based on the --url CLI option. If the
    # user did not give a recognized value, this raises an error
    output_scenario_name = context.output_scenario + "_nexus"
    output_model_name = context.output_model

    # Clone and build
    sc_ref = context.get_scenario()
    scen = sc_ref.clone(
        model=output_model_name, scenario=output_scenario_name, keep_solution=False
    )
    log.info(
        f" clone from {sc_ref.model}.{sc_ref.scenario} to {scen.model}.{scen.scenario}"
    )
    # Exporting the built model (Scenario) to GAMS with an optional case name
    caseName = scen.model + "__" + scen.scenario + "__v" + str(scen.version)

    # Build
    build(context, scen)

    # Set scenario as default
    scen.set_as_default()

    # Solve
    if macro:
        scen.solve(
            model="MESSAGE-MACRO",
            solve_options={"lpmethod": "4", "scaind": "1"},
            case=caseName,
        )
    else:
        scen.solve(solve_options={"lpmethod": "4"}, case=caseName)

    # if options["report"]:
    #     # Also output diagnostic reports
    #     from message_data.model.water import report, run_old_reporting
    #
    #     old reporting
    #     run_old_reporting(scen)

    #     log.info(f"Report plots to {rep.graph['config']['output_dir']}")

    #     mark_time()


@cli.command("cooling")
@common_params("regions")
@click.option("--rcps", type=click.Choice(_RCPS))
@click.option("--rels", type=click.Choice(_REL))
@click.pass_obj
def cooling_cli(context, regions, rcps, rels):
    """Build and solve model with new cooling technologies."""
    cooling(context, regions, rcps, rels)


def cooling(context, regions, rcps, rels):
    """Build and solve model with new cooling technologies.

    Use the --url option to specify the base scenario.

    Parameters
    ----------
    context : `class`:message.Context
        Information about target Scenario.
    regions : str (if not defined already in context.regions)
        Specifies what region definition is used ['R11','R12','ISO3']
    RCP : str
        Specifies the climate scenario used ['no_climate','6p0','2p6']

    """
    context.nexus_set = "cooling"
    context.RCP = rcps
    context.REL = rels

    from .build import main as build

    # Determine the output scenario name based on the --url CLI option. If the
    # user did not give a recognized value, this raises an error.

    output_scenario_name = context.output_scenario + "_cooling"
    output_model_name = context.output_model

    # Clone and build
    scen = context.get_scenario().clone(
        model=output_model_name, scenario=output_scenario_name, keep_solution=False
    )

    print(scen.model)
    print(scen.scenario)

    # Exporting the built model (Scenario) to GAMS with an optional case name
    caseName = scen.model + "__" + scen.scenario + "__v" + str(scen.version)

    # Build
    build(context, scen)

    # Solve
    scen.solve(solve_options={"lpmethod": "4"}, case=caseName)


@cli.command("report")
@click.pass_obj
@click.option(
    "--sdgs",
    default="baseline",
    help="Defines if and what water SDG measures are activated",
)
@click.option(
    "--water",
    is_flag=True,
    help="Default running legacy and water (full) otherwise only water, if specified",
)
@common_params("output_model")
def report_cli(context, output_model, sdgs, water=False):
    """function to run the water report_full from cli to the
    scenario defined by the user with --url

    Parameters
    ----------
    context : `class`:message.Context
        Information about target Scenario.
    output_model : str (optional, otherwise default args used)
        Specifies the model name of the scenarios which are run.
    SDG : Str
        Defines if and what water SDG measures are activated
    """
    reg = context.regions
    sc = context.get_scenario()
    if water:
        from message_ix_models.model.water.reporting import report

        report(sc, reg, sdgs)
    else:
        from message_ix_models.model.water.reporting import report_full

        report_full(sc, reg, sdgs)
